import importlib
import inspect
import json
import os
import pathlib
import pkgutil
from concurrent.futures import Future, ThreadPoolExecutor
from typing import List, Optional

from platypush.backend import Backend
from platypush.common.db import override_definitions
from platypush.common.reflection import Integration, Message as MessageMetadata
from platypush.config import Config
from platypush.plugins import Plugin, action
from platypush.message import Message
from platypush.message.event import Event
from platypush.message.response import Response
from platypush.utils import get_enabled_backends, get_enabled_plugins
from platypush.utils.mock import auto_mocks
from platypush.utils.manifest import Manifest, Manifests, PackageManagers

from ._cache import Cache
from ._serialize import ProcedureEncoder


class InspectPlugin(Plugin):
    """
    This plugin can be used to inspect platypush plugins and backends
    """

    _num_workers = 8
    """Number of threads to use for the inspection."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._cache = Cache()
        self._load_cache()

    @property
    def cache_file(self) -> str:
        """
        :return: The path to the components cache file.
        """
        import platypush

        return os.path.join(
            os.path.dirname(inspect.getfile(platypush)),
            'components.json.gz',
        )

    def _load_cache(self):
        """
        Loads the components cache from disk.
        """
        with self._cache.lock(), auto_mocks(), override_definitions():
            try:
                self._cache = Cache.load(self.cache_file)
            except Exception as e:
                self.logger.warning(
                    'Could not initialize the components cache from %s: %s',
                    self.cache_file,
                    e,
                )

                self.logger.exception(e)

    def refresh_cache(self, force: bool = False):
        """
        Refreshes the components cache.
        """
        cache_version_differs = self._cache.version != Cache.cur_version
        force = force or cache_version_differs

        with self._cache.lock(), auto_mocks(), override_definitions(), ThreadPoolExecutor(
            self._num_workers
        ) as pool:
            futures = []

            for base_type in [Plugin, Backend]:
                futures.append(
                    pool.submit(
                        self._scan_integrations,
                        base_type,
                        pool=pool,
                        force_refresh=force,
                        futures=futures,
                    )
                )

            for base_type in [Event, Response]:
                futures.append(
                    pool.submit(
                        self._scan_modules,
                        base_type,
                        pool=pool,
                        force_refresh=force,
                        futures=futures,
                    )
                )

            while futures:
                futures.pop().result()

        if self._cache.has_changes:
            self.logger.info('Saving new components cache to %s', self.cache_file)
            self._cache.dump(self.cache_file)
            self._cache.loaded_at = self._cache.saved_at

        return self._cache

    def _scan_integration(self, manifest: Manifest):
        """
        Scans a single integration from the manifest and adds it to the cache.
        """
        try:
            self._cache_integration(Integration.from_manifest(manifest.file))
        except Exception as e:
            self.logger.warning(
                'Could not import module %s: %s',
                manifest.package,
                e,
            )

    def _scan_integrations(
        self,
        base_type: type,
        pool: ThreadPoolExecutor,
        futures: List[Future],
        force_refresh: bool = False,
    ):
        """
        Scans the integrations with a manifest file (plugins and backends) and
        refreshes the cache.
        """
        for manifest in Manifests.by_base_class(base_type):
            # An integration metadata needs to be refreshed if it's been
            # modified since it was last loaded, or if it's not in the
            # cache.
            if force_refresh or self._needs_refresh(manifest.file):
                futures.append(pool.submit(self._scan_integration, manifest))

    def _scan_module(self, base_type: type, modname: str):
        """
        Scans a single module for objects that match the given base_type and
        adds them to the cache.
        """
        try:
            module = importlib.import_module(modname)
        except Exception as e:
            self.logger.warning('Could not import module %s: %s', modname, e)
            return

        for _, obj_type in inspect.getmembers(module):
            if (
                inspect.isclass(obj_type)
                and issubclass(obj_type, base_type)
                # Exclude the base_type itself
                and obj_type != base_type
            ):
                self.logger.info(
                    'Scanned %s: %s',
                    base_type.__name__,
                    f'{module.__name__}.{obj_type.__name__}',
                )

                self._cache.set(
                    base_type, obj_type, MessageMetadata.by_type(obj_type).to_dict()
                )

    def _scan_modules(
        self,
        base_type: type,
        pool: ThreadPoolExecutor,
        futures: List[Future],
        force_refresh: bool = False,
    ):
        """
        A generator that scans the modules given a ``base_type`` (e.g. ``Event``).

        It's a bit more inefficient than :meth:`._scan_integrations` because it
        needs to inspect all the members of a module to find the ones that
        match the given ``base_type``, but it works fine for simple components
        (like messages) that don't require extra recursive parsing and don't
        have a manifest.
        """
        prefix = base_type.__module__ + '.'
        path = str(pathlib.Path(inspect.getfile(base_type)).parent)

        for _, modname, __ in pkgutil.walk_packages(
            path=[path], prefix=prefix, onerror=lambda _: None
        ):
            try:
                filename = self._module_filename(path, '.'.join(modname.split('.')[3:]))
                if not (force_refresh or self._needs_refresh(filename)):
                    continue
            except Exception as e:
                self.logger.warning('Could not scan module %s: %s', modname, e)
                continue

            futures.append(pool.submit(self._scan_module, base_type, modname))

    def _needs_refresh(self, filename: str) -> bool:
        """
        :return: True if the given file needs to be refreshed in the cache.
        """
        dirname = os.path.dirname(filename)
        if not os.path.isdir(dirname):
            return True

        return os.lstat(dirname).st_mtime > (self._cache.saved_at or 0)

    @staticmethod
    def _module_filename(path: str, modname: str) -> str:
        """
        :param path: Path to the module.
        :param modname: Module name.
        :return: The full path to the module file.
        """
        filename = os.path.join(path, *modname.split('.')) + '.py'

        if not os.path.isfile(filename):
            filename = os.path.join(path, *modname.split('.'), '__init__.py')

        assert os.path.isfile(filename), f'No such file or directory: {filename}'
        return filename

    def _cache_integration(self, integration: Integration) -> dict:
        """
        :param integration: The :class:`.IntegrationMetadata` object.
        :return: The initialized component's metadata dict.
        """
        self.logger.info(
            'Scanned %s: %s', integration.base_type.__name__, integration.name
        )
        meta = integration.to_dict()
        self._cache.set(integration.base_type, integration.type, meta)
        return meta

    @action
    def get_all_plugins(self):
        """
        Get information about all the available plugins.
        """
        return json.dumps(self._cache.to_dict().get('plugins', {}), cls=Message.Encoder)

    @action
    def get_all_backends(self):
        """
        Get information about all the available backends.
        """
        return json.dumps(
            self._cache.to_dict().get('backends', {}), cls=Message.Encoder
        )

    @action
    def get_all_events(self):
        """
        Get information about all the available events.
        """
        return json.dumps(self._cache.to_dict().get('events', {}), cls=Message.Encoder)

    @action
    def get_all_responses(self):
        """
        Get information about all the available responses.
        """
        return json.dumps(
            self._cache.to_dict().get('responses', {}), cls=Message.Encoder
        )

    @action
    def get_procedures(self) -> dict:
        """
        Get the list of procedures installed on the device.
        """
        return json.loads(json.dumps(Config.get_procedures(), cls=ProcedureEncoder))

    @action
    def get_config(self, entry: Optional[str] = None) -> Optional[dict]:
        """
        Return the configuration of the application or of a section.

        :param entry: [Optional] configuration entry name to retrieve (e.g. ``workdir`` or ``backend.http``).
        :return: The requested configuration object.
        """
        if entry:
            return Config.get(entry)

        return Config.get()

    @action
    def get_enabled_plugins(self) -> List[str]:
        """
        Get the list of enabled plugins.
        """
        return list(get_enabled_plugins().keys())

    @action
    def get_enabled_backends(self) -> List[str]:
        """
        Get the list of enabled backends.
        """
        return list(get_enabled_backends().keys())

    @action
    def get_pkg_managers(self) -> dict:
        """
        Get the list of supported package managers. This is supposed to be an
        internal-only method, only used by the UI to populate the install
        commands.
        """
        pkg_manager = PackageManagers.scan()
        return {
            'items': {
                pkg.value.executable: {
                    'executable': pkg.value.executable,
                    'install': pkg.value.install,
                    'install_doc': pkg.value.install_doc,
                    'uninstall': pkg.value.uninstall,
                    'list': pkg.value.list,
                    'default_os': pkg.value.default_os,
                }
                for pkg in PackageManagers
            },
            'current': pkg_manager.value.executable if pkg_manager else None,
        }


# vim:sw=4:ts=4:et:



import math
import threading
from dataclasses import asdict, dataclass
from typing import Optional, Union, Tuple, Set

from platypush.plugins.camera.model.writer import (
    StreamWriter,
    VideoWriter,
    FileVideoWriter,
)
from platypush.plugins.camera.model.writer.preview import PreviewWriter


@dataclass
class CameraInfo:
    device: Optional[Union[int, str]]
    bind_address: Optional[str] = None
    capture_timeout: float = 0
    color_transform: Optional[Union[int, str]] = None
    ffmpeg_bin: Optional[str] = None
    fps: Optional[float] = None
    frames_dir: Optional[str] = None
    grayscale: Optional[bool] = None
    horizontal_flip: bool = False
    input_codec: Optional[str] = None
    input_format: Optional[str] = None
    listen_port: Optional[int] = None
    output_codec: Optional[str] = None
    output_format: Optional[str] = None
    resolution: Optional[Tuple[int, int]] = None
    rotate: Optional[float] = None
    scale_x: Optional[float] = None
    scale_y: Optional[float] = None
    stream_format: Optional[str] = None
    vertical_flip: bool = False
    warmup_frames: int = 0
    warmup_seconds: float = 0

    def set(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)

    def clone(self):
        return self.__class__(**asdict(self))


@dataclass
class Camera:
    info: CameraInfo
    start_event: threading.Event = threading.Event()
    stream_event: threading.Event = threading.Event()
    stop_stream_event: threading.Event = threading.Event()
    capture_thread: Optional[threading.Thread] = None
    stream_thread: Optional[threading.Thread] = None
    object = None
    stream: Optional[StreamWriter] = None
    preview: Optional[PreviewWriter] = None
    file_writer: Optional[FileVideoWriter] = None

    def get_outputs(self) -> Set[VideoWriter]:
        writers = set()
        # if self.preview and self.preview.is_alive():
        if self.preview and not self.preview.closed:
            writers.add(self.preview)

        if self.stream and not self.stream.closed:
            writers.add(self.stream)

        if self.file_writer and not self.file_writer.closed:
            writers.add(self.file_writer)

        return writers

    def effective_resolution(self) -> Tuple[int, int]:
        """
        Calculates the effective resolution of the camera in pixels, taking
        into account the base resolution, the scale and the rotation.
        """
        import numpy as np

        assert self.info.resolution, 'No base resolution specified'
        rot = (self.info.rotate or 0) * math.pi / 180
        sin = math.sin(rot)
        cos = math.cos(rot)
        scale = np.array([[self.info.scale_x or 1.0, self.info.scale_y or 1.0]])
        resolution = np.array([[self.info.resolution[0], self.info.resolution[1]]])
        rot_matrix = np.array([[sin, cos], [cos, sin]])
        resolution = (scale * abs(np.cross(rot_matrix, resolution)))[0]
        return int(round(resolution[0])), int(round(resolution[1]))


# vim:sw=4:ts=4:et:



from logging import getLogger
from queue import Full, Queue
from threading import Event
from time import time
from typing import Optional

import sounddevice as sd

from platypush.utils import wait_for_either

from ._state import AudioFrame, PauseState


class AudioRecorder:
    """
    Audio recorder component that uses the sounddevice library to record audio
    from the microphone.
    """

    def __init__(
        self,
        stop_event: Event,
        sample_rate: int,
        frame_size: int,
        channels: int,
        paused: bool = False,
        dtype: str = 'int16',
        queue_size: int = 100,
    ):
        self.logger = getLogger(__name__)
        self._audio_queue: Queue[AudioFrame] = Queue(maxsize=queue_size)
        self.frame_size = frame_size
        self._stop_event = Event()
        self._upstream_stop_event = stop_event
        self._paused_state = PauseState()
        if paused:
            self._paused_state.pause()
        else:
            self._paused_state.resume()

        self.stream = sd.InputStream(
            samplerate=sample_rate,
            channels=channels,
            dtype=dtype,
            blocksize=frame_size,
            callback=self._audio_callback,
        )

    @property
    def paused(self):
        return self._paused_state.paused

    def __enter__(self):
        """
        Start the audio stream.
        """
        return self.start()

    def __exit__(self, *_):
        """
        Stop the audio stream.
        """
        self.stop()

    def _audio_callback(self, indata, *_):
        if self.should_stop() or self.paused:
            return

        try:
            self._audio_queue.put_nowait(AudioFrame(indata.reshape(-1), time()))
        except Full:
            self.logger.warning('Audio queue is full, dropping audio frame')

    def read(self, timeout: Optional[float] = None):
        """
        Read an audio frame from the queue.

        :param timeout: Timeout in seconds. If None, the method will block until
            an audio frame is available.
        :return: Audio frame or None if the timeout has expired.
        """
        try:
            return self._audio_queue.get(timeout=timeout)
        except TimeoutError:
            self.logger.debug('Audio queue is empty')
            return None

    def start(self):
        """
        Start the audio stream.
        """
        self._stop_event.clear()
        self.stream.start()
        return self

    def stop(self):
        """
        Stop the audio stream.
        """
        self._stop_event.set()
        self.stream.stop()

    def pause(self):
        """
        Pause the audio stream.
        """
        self._paused_state.pause()

    def resume(self):
        """
        Resume the audio stream.
        """
        self._paused_state.resume()

    def toggle(self):
        """
        Toggle the audio stream pause state.
        """
        self._paused_state.toggle()

    def should_stop(self):
        return self._stop_event.is_set() or self._upstream_stop_event.is_set()

    def wait(self, timeout: Optional[float] = None):
        """
        Wait until the audio stream is stopped.
        """
        wait_for_either(self._stop_event, self._upstream_stop_event, timeout=timeout)

    def wait_start(self, timeout: Optional[float] = None):
        """
        Wait until the audio stream is started.
        """
        wait_for_either(
            self._stop_event,
            self._upstream_stop_event,
            self._paused_state._recording_event,  # pylint: disable=protected-access
            timeout=timeout,
        )



import json
import logging
import time

from platypush.message import Message


class Response(Message):
    """Response message class"""

    def __init__(
        self,
        target=None,
        origin=None,
        id=None,
        output=None,
        errors=None,
        timestamp=None,
        logging_level=logging.INFO,
    ):
        """
        :param target: Target
        :type target: str
        :param origin: Origin
        :type origin: str
        :param output: Output
        :param errors: Errors
        :param id: Message ID this response refers to
        :type id: str
        :param timestamp: Message timestamp
        :type timestamp: float
        """

        super().__init__(timestamp=timestamp, logging_level=logging_level)
        self.target = target
        self.output = self._parse_msg(output)
        self.errors = self._parse_msg(errors or [])
        self.origin = origin
        self.id = id
        self._logger = logging.getLogger('platypush:responses')

    def is_error(self):
        """Returns True if the response has errors"""
        return len(self.errors) != 0

    @classmethod
    def _parse_msg(cls, msg):
        if isinstance(msg, (bytes, bytearray)):
            msg = msg.decode('utf-8')
        if isinstance(msg, str):
            try:
                msg = json.loads(msg.strip())
            except ValueError:
                pass

        return msg

    @classmethod
    def build(cls, msg):
        msg = super().parse(msg)
        args = {
            'target': msg['target'],
            'output': msg['response']['output'],
            'errors': msg['response']['errors'],
            'timestamp': msg['_timestamp'] if '_timestamp' in msg else time.time(),
            'logging_level': msg.get('_logging_level', logging.INFO),
        }

        if 'id' in msg:
            args['id'] = msg['id']
        if 'origin' in msg:
            args['origin'] = msg['origin']

        return cls(**args)

    def __str__(self):
        """
        Overrides the ``str()`` operator and converts
        the message into a UTF-8 JSON string
        """
        output = (
            self.output if self.output is not None else {'success': not self.errors}
        )

        response_dict = {
            'id': self.id,
            'type': 'response',
            'target': self.target if hasattr(self, 'target') else None,
            'origin': self.origin if hasattr(self, 'origin') else None,
            '_timestamp': self.timestamp,
            'response': {
                'output': output,
                'errors': self.errors,
            },
        }

        if self.logging_level:
            response_dict['_logging_level'] = self.logging_level

        return json.dumps(response_dict, cls=self.Encoder)

    def log(self, *args, **kwargs):
        self.logging_level = logging.WARNING if self.is_error() else logging.INFO
        super().log(*args, **kwargs)


# vim:sw=4:ts=4:et:



import os

from abc import ABC, abstractmethod
from datetime import datetime
from email.message import Message
from email.mime.application import MIMEApplication
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from mimetypes import guess_type
from typing import Dict, Optional, Sequence, Union

from dateutil import tz

from .._utils import normalize_from_header
from ._base import BaseMailPlugin


class MailOutPlugin(BaseMailPlugin, ABC):
    """
    Base class for mail out plugins.
    """

    @abstractmethod
    def send_message(self, message: Message, **_):
        raise NotImplementedError()

    @staticmethod
    def _file_to_part(file: str) -> MIMEBase:
        _type, _subtype, _type_class = 'application', 'octet-stream', MIMEApplication
        mime_type, _sub_subtype = guess_type(file)

        if mime_type:
            _type, _subtype = mime_type.split('/')
        if _sub_subtype:
            _subtype += ';' + _sub_subtype

        if _type == 'application':
            _type_class = MIMEApplication
        elif _type == 'audio':
            _type_class = MIMEAudio
        elif _type == 'image':
            _type_class = MIMEImage
        elif _type == 'text':
            _type_class = MIMEText

        args = {}
        if _type_class != MIMEText:
            mode = 'rb'
            args['Name'] = os.path.basename(file)
        else:
            mode = 'r'

        with open(file, mode) as f:
            return _type_class(f.read(), _subtype, **args)

    @classmethod
    def create_message(
        cls,
        to: Union[str, Sequence[str]],
        from_: Optional[str] = None,
        cc: Optional[Union[str, Sequence[str]]] = None,
        bcc: Optional[Union[str, Sequence[str]]] = None,
        subject: str = '',
        body: str = '',
        body_type: str = 'plain',
        attachments: Optional[Sequence[str]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> Message:
        assert from_, 'from/from_ field not specified'

        content = MIMEText(body, body_type)
        if attachments:
            msg = MIMEMultipart()
            msg.attach(content)

            for attachment in attachments:
                attachment = os.path.abspath(os.path.expanduser(attachment))
                assert os.path.isfile(attachment), f'No such file: {attachment}'
                part = cls._file_to_part(attachment)
                part[
                    'Content-Disposition'
                ] = f'attachment; filename="{os.path.basename(attachment)}"'
                msg.attach(part)
        else:
            msg = content

        msg['From'] = from_
        msg['To'] = to if isinstance(to, str) else ', '.join(to)
        msg['Cc'] = ', '.join(cc) if cc else ''
        msg['Bcc'] = ', '.join(bcc) if bcc else ''
        msg['Subject'] = subject
        msg['Date'] = (
            datetime.now()
            .replace(tzinfo=tz.tzlocal())
            .strftime('%a, %d %b %Y %H:%M:%S %z')
        )

        if headers:
            for name, value in headers.items():
                msg.add_header(name, value)

        return msg

    def send(
        self,
        to: Union[str, Sequence[str]],
        from_: Optional[str] = None,
        cc: Optional[Union[str, Sequence[str]]] = None,
        bcc: Optional[Union[str, Sequence[str]]] = None,
        subject: str = '',
        body: str = '',
        body_type: str = 'plain',
        attachments: Optional[Sequence[str]] = None,
        headers: Optional[Dict[str, str]] = None,
        **args,
    ):
        if not from_ and 'from' in args:
            from_ = args.pop('from')

        msg = self.create_message(
            to=to,
            from_=normalize_from_header(from_, self.account, self.server),
            cc=cc,
            bcc=bcc,
            subject=subject,
            body=body,
            body_type=body_type,
            attachments=attachments,
            headers=headers,
        )

        return self.send_message(msg, **args)


# vim:sw=4:ts=4:et:



from dataclasses import asdict
import json
from logging import getLogger
import pathlib
from threading import Event, RLock, Timer
from typing import Final, Optional

from .._mixins import XmppConfigMixin
from ._model import SerializedState, XmppState


class StateSerializer(XmppConfigMixin):
    """
    Serializes to file the state of the client upon new events through a
    timer-based mechanism.
    """

    _DEFAULT_FLUSH_TIMEOUT: Final[float] = 2
    _EMPTY_STATE: Final[SerializedState] = SerializedState()

    def __init__(self, *args, flush_timeout: float = _DEFAULT_FLUSH_TIMEOUT, **kwargs):
        """
        :param flush_timeout: How long the scheduler should wait before
            flushing the state.
        """
        super().__init__(*args, **kwargs)
        self.flush_timeout = flush_timeout
        self._timer: Optional[Timer] = None
        self._state_lock: Final[RLock] = RLock()
        self._state: Optional[XmppState] = None
        self._flush_scheduled: Final[Event] = Event()
        self.logger = getLogger(__name__)

    def _writer_inner(self, filename: str):
        if not self._state:
            return

        self.logger.debug("Serializing state to file: %s", filename)
        pathlib.Path(filename).parent.mkdir(parents=True, exist_ok=True)
        with open(filename, "w") as f:
            json.dump(asdict(self._state.serialize()), f)

    def _writer(self):
        """
        Write the current state to the file.
        """

        state_file = self._config.state_file
        if not state_file:
            return

        with self._state_lock:
            try:
                self._writer_inner(state_file)
            finally:
                self._reset()

    def _reset(self):
        """
        Reset the timer state after normal termination, error or cancellation.
        """
        self._flush_scheduled.clear()
        self._timer = None

    def load(self) -> SerializedState:
        """
        :return: The previous state read from the configured state file.
        """
        state_file = self._config.state_file
        if not (state_file and self._config.restore_state):
            return self._EMPTY_STATE

        try:
            with open(state_file, "r") as f:
                return SerializedState.load(json.load(f))
        except FileNotFoundError:
            self.logger.info("No previous state file found at %s", state_file)
            return self._EMPTY_STATE
        except ValueError:
            self.logger.warning(
                "Invalid or corrupt state file found at %s, it will be reset",
                state_file,
            )
            return self._EMPTY_STATE

    def enqueue(self, state: XmppState):
        """
        Schedule an update of the stored state.
        """
        with self._state_lock:
            self._state = state

            if not self.is_pending():
                self.logger.debug(
                    "Serialization writer scheduled in %f seconds", self.flush_timeout
                )
                self._timer = Timer(self.flush_timeout, self._writer)
                self._timer.name = "xmpp:StateSerializer"
                self._timer.start()

            self._flush_scheduled.set()

    def flush(self):
        """
        Flush the state immediately, without waiting for the next schedule.
        """
        with self._state_lock:
            self._writer()

    def is_pending(self) -> bool:
        """
        :return: ``True`` if there is a pending serialization task, ``False``
            otherwise.
        """
        return self._timer is not None and self._flush_scheduled.is_set()

    def wait(self, timeout: Optional[float] = None):
        """
        If a serialization task is pending or running, wait for it to terminate.
        """
        if self._timer and self.is_pending():
            self._timer.join(timeout)

            with self._state_lock:
                if self._timer and self.is_pending():
                    self.logger.warning(
                        "The state serialization task did not terminate in time"
                    )

                    self.cancel()

    def cancel(self):
        """
        Cancel the timer, if it is running.
        """
        if self._timer:
            self._timer.cancel()

        self._reset()



from dataclasses import dataclass
from typing import Optional

from ._common import DEFAULT_TIMEOUT


@dataclass
class MopidyConfig:
    """
    Mopidy configuration.
    """

    host: str = 'localhost'
    port: int = 6680
    ssl: bool = False
    timeout: Optional[float] = DEFAULT_TIMEOUT

    @property
    def url(self) -> str:
        return f'ws{"s" if self.ssl else ""}://{self.host}:{self.port}/mopidy/ws'


# vim:sw=4:ts=4:et:



class MopidyException(Exception):
    """
    Base class for all Mopidy exceptions.
    """


class EmptyTrackException(MopidyException, ValueError):
    """
    Raised when a parsed track is empty.
    """



from ._model import (
    MajorDeviceClass,
    MajorServiceClass,
    MinorDeviceClass,
    Protocol,
    ServiceClass,
)
from ._types import RawServiceClass

__all__ = [
    "MajorDeviceClass",
    "MajorServiceClass",
    "MinorDeviceClass",
    "Protocol",
    "RawServiceClass",
    "ServiceClass",
]



import urllib.parse
from typing import Optional

from platypush.context import get_plugin
from platypush.plugins import Plugin, action


class TtsPlugin(Plugin):
    """
    Default Text-to-Speech plugin. It leverages Google Translate's unofficial
    frontend API.
    """

    def __init__(self, language='en-US', **player_args):
        """
        :param language: Language code (default: ``en-US``).
        :param player_args: Additional arguments to be passed to
            :meth:`platypush.plugins.sound.SoundPlugin.play` (like volume,
            duration, channels etc.).
        """
        super().__init__()
        self.language = language
        self.player_args = player_args or {}

    def _playback(self, resource: str, **kwargs):
        audio = get_plugin('sound')
        assert audio
        audio.play(resource, **{**self.player_args, **kwargs})

    @action
    def say(
        self,
        text: str,
        language: Optional[str] = None,
        **player_args,
    ):
        """
        Say some text.

        :param text: Text to say.
        :param language: Language code override.
        :param player_args: Extends the additional arguments to be passed to
            :meth:`platypush.plugins.sound.SoundPlugin.play` (like volume,
            duration, channels etc.).
        """
        language = language or self.language
        url = 'https://translate.google.com/translate_tts?' + urllib.parse.urlencode(
            {
                'ie': 'UTF-8',
                'client': 'tw-ob',
                'tl': language,
                'q': text,
            }
        )

        self._playback(url, **player_args)

    @action
    def stop(self):
        """
        Stop the playback.
        """
        try:
            audio = get_plugin('sound')
        except Exception:
            return

        if audio:
            audio.stop_playback()


# vim:sw=4:ts=4:et:



import math
import re
from typing import Optional, Union


class SoundParser:
    """
    A utility mixin with some methods to parse and convert sound information -
    e.g. MIDI notes from strings, MIDI notes to frequencies, and the other way
    around.
    """

    _DEFAULT_A4_FREQUENCY = 440.0
    _MIDI_NOTE_REGEX = re.compile(r'^([A-G])([#b]?)(-?[0-9]+)$')
    _MID_A_MIDI_NOTE = 69
    _NOTE_OFFSETS = {
        'C': 0,
        'C#': 1,
        'Db': 1,
        'D': 2,
        'D#': 3,
        'Eb': 3,
        'E': 4,
        'F': 5,
        'F#': 6,
        'Gb': 6,
        'G': 7,
        'G#': 8,
        'Ab': 8,
        'A': 9,
        'A#': 10,
        'Bb': 10,
        'B': 11,
    }

    _ALTERATION_OFFSETS = {
        'b': -1,
        '': 0,
        '#': 1,
    }

    def __init__(self, *_, ref_frequency: float = _DEFAULT_A4_FREQUENCY, **__) -> None:
        self._ref_frequency = ref_frequency

    @staticmethod
    def _get_alteration_offset(alt: str) -> int:
        """
        Calculate the MIDI note offset given by its reported sharp/flat alteration.
        """
        if alt == '#':
            return 1
        if alt == 'b':
            return -1
        return 0

    @classmethod
    def get_midi_note(cls, note: Union[str, int]) -> int:
        """
        Convert a MIDI note given as input (either an integer or a string like
        'C4') to a MIDI note number.

        :raise: ValueError
        """

        if isinstance(note, str):
            note = note[:1].upper() + note[1:]
            m = cls._MIDI_NOTE_REGEX.match(note)
            if not m:
                raise ValueError(f'Invalid MIDI note: {note}')

            base_note, alteration, octave = m.groups()
            octave = int(octave)
            note_offset = cls._NOTE_OFFSETS[base_note] + cls._get_alteration_offset(
                alteration
            )

            octave_offset = (octave + 1) * 12
            note = octave_offset + note_offset

        if isinstance(note, int):
            if not 0 <= note <= 127:
                raise ValueError(f'MIDI note out of range: {note}')
            return note

        raise ValueError(f'Invalid MIDI note: {note}')

    def note_to_freq(
        self, midi_note: Union[int, str], ref_frequency: Optional[float] = None
    ):
        """
        Converts a MIDI note to its frequency in Hz

        :param midi_note: MIDI note to convert
        :param ref_frequency: Reference A4 frequency override (default: 440 Hz).
        """

        note = self.get_midi_note(midi_note)
        return (2.0 ** ((note - self._MID_A_MIDI_NOTE) / 12.0)) * (
            ref_frequency or self._ref_frequency
        )

    def freq_to_note(self, frequency: float, ref_frequency: Optional[float] = None):
        """
        Converts a frequency in Hz to its closest MIDI note

        :param frequency: Frequency in Hz
        :param ref_frequency: Reference A4 frequency override (default: 440 Hz).
        """

        std_freq = ref_frequency or self._ref_frequency
        return int(12.0 * math.log(frequency / std_freq, 2) + self._MID_A_MIDI_NOTE)



from marshmallow_dataclass import class_schema

from ._base import NetworkInterfaceBaseSchema
from ._model import NetworkInterface


NetworkInterfaceSchema = class_schema(
    NetworkInterface, base_schema=NetworkInterfaceBaseSchema
)



from marshmallow import EXCLUDE, fields
from marshmallow.schema import Schema


class PrinterSchema(Schema):
    """
    Schema for the printers returned by the CUPS API.
    """

    class Meta:  # type: ignore
        """
        Exclude unknown fields.
        """

        unknown = EXCLUDE

    name = fields.String(
        required=True,
        metadata={
            'description': 'The name of the printer.',
            'example': 'HP_DeskJet_5820_series_585C26',
        },
    )

    printer_type = fields.String(
        required=True,
        data_key='printer-type',
        metadata={
            'description': 'Unique type ID of the printer.',
            'example': 2101260,
        },
    )

    info = fields.String(
        required=True,
        data_key='printer-info',
        metadata={
            'description': 'Human-readable description of the printer.',
            'example': 'HP DeskJet 5820 series',
        },
    )

    uri = fields.String(
        required=True,
        data_key='device-uri',
        metadata={
            'description': 'The URI of the printer.',
            'example': (
                'dnssd://HP%20DeskJet%205820%20series%20%5B585C26%5D._ipp._tcp.local/'
                '?uuid=1c852a4d-b800-1f08-abcd-705a0f585c26'
            ),
        },
    )

    state = fields.String(
        required=True,
        data_key='printer-state',
        metadata={
            'description': 'The state of the printer.',
            'example': 3,
        },
    )

    is_shared = fields.Boolean(
        required=True,
        data_key='printer-is-shared',
        metadata={
            'description': 'Whether the printer is shared.',
            'example': False,
        },
    )

    state_message = fields.String(
        required=True,
        data_key='printer-state-message',
        metadata={
            'description': 'The state message of the printer.',
            'example': 'Idle',
        },
    )

    state_reasons = fields.List(
        fields.String(),
        required=True,
        data_key='printer-state-reasons',
        metadata={
            'description': 'The reasons for the printer state.',
            'example': ['none'],
        },
    )

    location = fields.String(
        required=True,
        data_key='printer-location',
        metadata={
            'description': 'Human-readable location of the printer.',
            'example': 'Living Room',
        },
    )

    uri_supported = fields.String(
        required=True,
        data_key='printer-uri-supported',
        metadata={
            'description': 'The URI of the printer exposed by the CUPS server.',
            'example': 'ipp://localhost:631/printers/HP_OfficeJet_5230',
        },
    )

    make_and_model = fields.String(
        required=True,
        data_key='printer-make-and-model',
        metadata={
            'description': 'The make and model of the printer.',
            'example': 'HP Officejet 5200 Series, hpcups 3.19.1',
        },
    )


class JobAddedSchema(Schema):
    """
    Schema for a printer job added event.
    """

    printer = fields.String(
        required=True,
        metadata={
            'description': 'The name of the printer.',
            'example': 'HP_DeskJet_5820_series_585C26',
        },
    )

    job_id = fields.Integer(
        required=True,
        data_key='job-id',
        metadata={
            'description': 'The ID of the job.',
            'example': 1,
        },
    )



import logging
import os
import re
from threading import RLock
from typing import Optional

import numpy as np
import pvorca
import sounddevice as sd

from platypush.config import Config
from platypush.plugins import action
from platypush.plugins.tts import TtsPlugin


class TextConversionUtils:
    """
    Utility class to convert text to a format that is supported by the Orca TTS
    engine.

    This pre-processing step is necessary until the issue is fixed:
    https://github.com/Picovoice/orca/issues/10.
    """

    _logger = logging.getLogger(__name__)
    _number_re = re.compile(r'(([0-9]+\.[0-9]+)|([0-9]+\,[0-9]+)|([0-9]+))')
    _conversions_map = {
        (re.compile(r'\s*[(){}\[\]<>]'), ', '),
        (re.compile(r'[;]'), '.'),
        (re.compile(r'[@#]'), ' at '),
        (re.compile(r'[$]'), ' dollar '),
        (re.compile(r'[%]'), ' percent '),
        (re.compile(r'[&]'), ' and '),
        (re.compile(r'[+]'), ' plus '),
        (re.compile(r'[=]'), ' equals '),
        (re.compile(r'[|]'), ' or '),
        (re.compile(r'[~]'), ' tilde '),
        (re.compile(r'[`]'), ''),
        (re.compile(r'[*]'), ' star '),
        (re.compile(r'[\\/]'), ' slash '),
        (re.compile(r'[_]'), '  underscore '),
        # Anything that isn't a letter or supported punctuation is replaced with a space
        (re.compile(r'[^a-zA-Z,.:?!\-\'" ]'), ' '),
    }

    @classmethod
    def _convert_digits(cls, text: str) -> str:
        try:
            from num2words import num2words
        except ImportError:
            cls._logger.warning('num2words is not installed, skipping digit conversion')
            return text

        while match := cls._number_re.search(text):
            number = match.group(1)
            text = text.replace(number, num2words(float(number.replace(',', ''))))

        return text

    @classmethod
    def convert(cls, text: str) -> str:
        text = cls._convert_digits(text)

        for pattern, replacement in TextConversionUtils._conversions_map:
            text = pattern.sub(replacement, text)

        return text


class TtsPicovoicePlugin(TtsPlugin):
    """
    This TTS plugin enables you to render text as audio using `Picovoice
    <https://picovoice.ai>`_'s (still experimental) `Orca TTS engine
    <https://github.com/Picovoice/orca>`_.

    Take a look at
    :class:`platypush.plugins.assistant.picovoice.AssistantPicovoicePlugin`
    for details on how to sign up for a Picovoice account and get the API key.

    Also note that using the TTS features requires you to select Orca from the
    list of products available for your account on the `Picovoice console
    <https://console.picovoice.ai>`_.
    """

    def __init__(
        self,
        access_key: Optional[str] = None,
        model_path: Optional[str] = None,
        **kwargs,
    ):
        """
        :param access_key: Picovoice access key. If it's not specified here,
            then it must be specified on the configuration of
            :class:`platypush.plugins.assistant.picovoice.AssistantPicovoicePlugin`.
        :param model_path: Path of the TTS model file (default: use the default
            English model).
        """
        super().__init__(**kwargs)
        if not access_key:
            access_key = Config.get('assistant.picovoice', {}).get('access_key')
            assert (
                access_key
            ), 'No access key specified and no assistant.picovoice plugin found'

        self.model_path = model_path
        self.access_key = access_key
        if model_path:
            model_path = os.path.expanduser(model_path)

        self._stream: Optional[sd.OutputStream] = None
        self._stream_lock = RLock()

    def _play_audio(self, orca: pvorca.Orca, pcm: np.ndarray):
        with self._stream_lock:
            self.stop()
            self._stream = sd.OutputStream(
                samplerate=orca.sample_rate,
                channels=1,
                dtype='int16',
            )

        try:
            self._stream.start()
            self._stream.write(pcm)
        except Exception as e:
            self.logger.warning('Error playing audio: %s: %s', type(e), str(e))
        finally:
            try:
                self.stop()
                self._stream.close()
            except Exception as e:
                self.logger.warning(
                    'Error stopping audio stream: %s: %s', type(e), str(e)
                )
            finally:
                if self._stream:
                    self._stream = None

    def get_orca(self, model_path: Optional[str] = None):
        if not model_path:
            model_path = self.model_path
        if model_path:
            model_path = os.path.expanduser(model_path)

        return pvorca.create(access_key=self.access_key, model_path=model_path)

    @action
    def say(
        self,
        text: str,
        *_,
        output_file: Optional[str] = None,
        speech_rate: Optional[float] = None,
        model_path: Optional[str] = None,
        **__,
    ):
        """
        Say some text.

        :param text: Text to say.
        :param output_file: If set, save the audio to the specified file.
            Otherwise play it.
        :param speech_rate: Speech rate (default: None).
        :param model_path: Path of the TTS model file (default: use the default
            configured model).
        """
        # This is a temporary workaround until this issue is fixed:
        # https://github.com/Picovoice/orca/issues/10.
        text = TextConversionUtils.convert(text)
        orca = self.get_orca(model_path=model_path)

        if output_file:
            orca.synthesize_to_file(
                text, os.path.expanduser(output_file), speech_rate=speech_rate
            )
            return

        self._play_audio(
            orca=orca,
            pcm=np.array(
                orca.synthesize(text, speech_rate=speech_rate)[0],
                dtype='int16',
            ),
        )

    @action
    def stop(self):
        """
        Stop the currently playing audio.
        """
        with self._stream_lock:
            if not self._stream:
                return

            self._stream.stop()


# vim:sw=4:ts=4:et:



import base64
import re
from functools import lru_cache
from typing import List, Optional

import requests

from platypush.plugins import Plugin, action
from platypush.schemas.piped import (
    PipedChannelSchema,
    PipedPlaylistSchema,
    PipedVideoSchema,
)


class YoutubePlugin(Plugin):
    r"""
    YouTube plugin.

    Unlike other Google plugins, this plugin doesn't rely on the Google API.

    That's because the official YouTube API has been subject to many changes to
    prevent scraping, and it requires the user to tinker with the OAuth layer,
    app permissions and app validation in order to get it working.

    Instead, it relies on a `Piped <https://docs.piped.video/>`_, an open-source
    alternative YouTube gateway.

    It thus requires a link to a valid Piped instance.
    """

    _timeout = 20

    def __init__(
        self,
        piped_api_url: str = 'https://pipedapi.kavin.rocks',
        auth_token: Optional[str] = None,
        **kwargs,
    ):
        """
        :param piped_api_url: Base API URL of the Piped instance (default:
            ``https://pipedapi.kavin.rocks``).
        :param auth_token: Optional authentication token from the Piped
            instance. This is required if you want to access your private feed
            and playlists, but not for searching public videos.

            In order to retrieve an authentication token:

              1. Login to your configured Piped instance.
              2. Copy the RSS/Atom feed URL on the _Feed_ tab.
              3. Copy the ``auth_token`` query parameter from the URL.
              4. Enter it in the ``auth_token`` field in the ``youtube`` section
                 of the configuration file.

        """
        super().__init__(**kwargs)
        self._piped_api_url = piped_api_url
        self._auth_token = auth_token

    def _api_url(self, path: str = '') -> str:
        return f"{self._piped_api_url}/{path}"

    def _request(
        self,
        path: str,
        method: str = 'get',
        body: Optional[str] = None,
        auth: bool = True,
        **kwargs,
    ):
        timeout = kwargs.pop('timeout', self._timeout)
        if auth:
            kwargs['params'] = kwargs.get('params', {})
            kwargs['params']['authToken'] = self._auth_token
            kwargs['headers'] = kwargs.get('headers', {})
            kwargs['headers']['Authorization'] = self._auth_token

        if body:
            kwargs['data'] = body

        func = getattr(requests, method.lower())
        rs = func(self._api_url(path), timeout=timeout, **kwargs)
        rs.raise_for_status()

        try:
            return rs.json()
        except (TypeError, ValueError):
            return {}

    @lru_cache(maxsize=10)  # noqa
    def _get_channel(self, id: str) -> dict:  # pylint: disable=redefined-builtin
        if (
            id.startswith('http')
            or id.startswith('https')
            or id.startswith('/channel/')
        ):
            id = id.split('/')[-1]

        return (
            PipedChannelSchema().dump(self._request(f'channel/{id}')) or {}  # type: ignore
        )

    @staticmethod
    def _get_video_id(id_or_url: str) -> str:
        m = re.search(r'/watch\?v=([^&]+)', id_or_url)
        if m:
            return m.group(1)

        return id_or_url

    @staticmethod
    def _dump_item(item: dict) -> dict:
        if item.get('type') == 'stream':
            return dict(PipedVideoSchema().dump(item))

        if item.get('type') == 'channel':
            return dict(PipedChannelSchema().dump(item))

        if item.get('type') == 'playlist':
            return dict(PipedPlaylistSchema().dump(item))

        return item

    @action
    def search(self, query: str, **_) -> List[dict]:
        """
        Search for YouTube videos.

        :param query: Query string.
        :return: .. schema:: piped.PipedVideoSchema(many=True)
        """
        self.logger.info('Searching YouTube for "%s"', query)
        rs = self._request('search', auth=False, params={'q': query, 'filter': 'all'})
        results = [self._dump_item(item) for item in rs.get('items', [])]
        self.logger.info(
            '%d YouTube results for the search query "%s"',
            len(results),
            query,
        )

        return results

    @action
    def get_feed(self) -> List[dict]:
        """
        Retrieve the YouTube feed.

        Depending on your account settings on the configured Piped instance,
        this may return either the latest videos uploaded by your subscribed
        channels, or the trending videos in the configured area.

        :return: .. schema:: piped.PipedVideoSchema(many=True)
        """
        return PipedVideoSchema(many=True).dump(self._request('feed')) or []

    @action
    def get_playlists(self) -> List[dict]:
        """
        Retrieve the playlists saved by the user logged in to the Piped
        instance.

        :return: .. schema:: piped.PipedPlaylistSchema(many=True)
        """
        return (
            PipedPlaylistSchema(many=True).dump(self._request('user/playlists')) or []
        )

    @action
    def get_playlist(self, id: str) -> List[dict]:  # pylint: disable=redefined-builtin
        """
        Retrieve the videos in a playlist.

        :param id: Piped playlist ID.
        :return: .. schema:: piped.PipedVideoSchema(many=True)
        """
        return (
            PipedVideoSchema(many=True).dump(
                self._request(f'playlists/{id}').get('relatedStreams', [])
            )
            or []
        )

    @action
    def get_subscriptions(self) -> List[dict]:
        """
        Retrieve the channels subscribed by the user logged in to the Piped
        instance.

        :return: .. schema:: piped.PipedChannelSchema(many=True)
        """
        return PipedChannelSchema(many=True).dump(self._request('subscriptions')) or []

    @action
    def get_channel(
        self,
        id: str,  # pylint: disable=redefined-builtin
        next_page_token: Optional[str] = None,
    ) -> dict:
        """
        Retrieve the information and videos of a channel given its ID or URL.

        :param id: Channel ID or URL.
        :param next_page_token: Optional token to retrieve the next page of
            results.
        :return: .. schema:: piped.PipedChannelSchema
        """
        if (
            id.startswith('http')
            or id.startswith('https')
            or id.startswith('/channel/')
        ):
            id = id.split('/')[-1]

        info = {}
        if next_page_token:
            info = self._get_channel(id).copy()
            info.pop('next_page_token', None)
            info['items'] = []
            next_page = base64.b64decode(next_page_token.encode()).decode()
            response = {
                **info,
                **self._request(
                    f'nextpage/channel/{id}', params={'nextpage': next_page}, auth=False
                ),
            }
        else:
            response = self._request(f'channel/{id}')

        return PipedChannelSchema().dump(response) or {}  # type: ignore

    @action
    def add_to_playlist(self, video_id: str, playlist_id: str):
        """
        Add a video to a playlist.

        :param video_id: YouTube video ID.
        :param playlist_id: Piped playlist ID.
        """
        self._request(
            'user/playlists/add',
            method='post',
            json={
                'videoIds': [self._get_video_id(video_id)],
                'playlistId': playlist_id,
            },
        )

    @action
    def remove_from_playlist(
        self,
        playlist_id: str,
        video_id: Optional[str] = None,
        index: Optional[int] = None,
    ):
        """
        Remove a video from a playlist.

        Note that either the video ID or the index must be provided.

        :param video_id: YouTube video ID or URL.
        :param index: (0-based) index of the video in the playlist.
        :param playlist_id: Piped playlist ID.
        """
        assert video_id or index, 'Either the video ID or the index must be provided'

        if index is None:
            assert video_id
            video_id = self._get_video_id(video_id)
            index = next(
                (
                    i
                    for i, v in enumerate(
                        self._request(f'playlists/{playlist_id}').get(
                            'relatedStreams', []
                        )
                    )
                    if self._get_video_id(v.get('url')) == video_id
                ),
                None,
            )

        if index is None:
            self.logger.warning(
                'Video %s not found in the playlist %s', video_id, playlist_id
            )
            return

        self._request(
            'user/playlists/remove',
            method='post',
            json={
                'index': index,
                'playlistId': playlist_id,
            },
        )

    @action
    def create_playlist(self, name: str) -> dict:
        """
        Create a new playlist.

        :param name: Playlist name.
        :return: Playlist information.
        """
        name = name.strip()
        assert name, 'Playlist name cannot be empty'

        playlist_id = self._request(
            'user/playlists/create',
            method='post',
            json={'name': name},
        ).get('playlistId')

        assert playlist_id, 'Failed to create the playlist'
        playlists = self._request('user/playlists')
        new_playlist = next((p for p in playlists if p.get('id') == playlist_id), None)

        assert new_playlist, 'Failed to retrieve the new playlist'
        return dict(PipedPlaylistSchema().dump(new_playlist) or {})

    @action
    def rename_playlist(
        self, id: str, name: Optional[str] = None, description: Optional[str] = None
    ):  # pylint: disable=redefined-builtin
        """
        Rename a playlist.

        :param id: Piped playlist ID.
        :param name: New playlist name.
        :param description: New playlist description.
        """
        args = {}
        if name:
            args['newName'] = name
        if description:
            args['description'] = description
        if not args:
            self.logger.info('No new name or description provided')
            return

        self._request(
            'user/playlists/rename',
            method='post',
            json={
                'playlistId': id,
                **args,
            },
        )

    @action
    def delete_playlist(self, id: str):  # pylint: disable=redefined-builtin
        """
        Delete a playlist.

        :param id: Piped playlist ID.
        """
        self._request(
            'user/playlists/delete',
            method='post',
            json={'playlistId': id},
        )

    @action
    def is_subscribed(self, channel_id: str) -> bool:
        """
        Check if the user is subscribed to a channel.

        :param channel_id: YouTube channel ID.
        :return: True if the user is subscribed to the channel, False otherwise.
        """
        return self._request(
            'subscribed',
            params={'channelId': channel_id},
        ).get('subscribed', False)

    @action
    def subscribe(self, channel_id: str):
        """
        Subscribe to a channel.

        :param channel_id: YouTube channel ID.
        """
        self._request(
            'subscribe',
            method='post',
            json={'channelId': channel_id},
        )

    @action
    def unsubscribe(self, channel_id: str):
        """
        Unsubscribe from a channel.

        :param channel_id: YouTube channel ID.
        """
        self._request(
            'unsubscribe',
            method='post',
            json={'channelId': channel_id},
        )


# vim:sw=4:ts=4:et:



from ._config import AccountConfig, ServerConfig
from ._mail import FolderStatus, Mail, MailFlagType, AccountsStatus
from ._transport import TransportEncryption


__all__ = [
    'AccountConfig',
    'FolderStatus',
    'Mail',
    'MailFlagType',
    'AccountsStatus',
    'ServerConfig',
    'TransportEncryption',
]



from dataclasses import dataclass, field
from typing import Optional

from platypush.schemas.dataclasses import percent_field


@dataclass
class Disk:
    """
    Disk data class.
    """

    device: str = field(
        metadata={
            'metadata': {
                'description': 'Path/identifier of the disk/partition',
                'example': '/dev/sda1',
            }
        }
    )

    mountpoint: Optional[str] = field(
        metadata={
            'metadata': {
                'description': 'Where the disk is mounted',
                'example': '/home',
            }
        }
    )

    fstype: Optional[str] = field(
        metadata={
            'metadata': {
                'description': 'Filesystem type',
                'example': 'ext4',
            }
        }
    )

    opts: Optional[str] = field(
        metadata={
            'metadata': {
                'description': 'Extra mount options passed to the partition',
                'example': 'rw,relatime,fmask=0022,dmask=0022,utf8',
            }
        }
    )

    total: int = field(
        metadata={
            'metadata': {
                'description': 'Total available space, in bytes',
            }
        }
    )

    used: int = field(
        metadata={
            'metadata': {
                'description': 'Used disk space, in bytes',
            }
        }
    )

    free: int = field(
        metadata={
            'metadata': {
                'description': 'Free disk space, in bytes',
            }
        }
    )

    read_count: Optional[int] = field(
        metadata={
            'metadata': {
                'description': 'Number of recorded read operations',
            }
        }
    )

    write_count: Optional[int] = field(
        metadata={
            'metadata': {
                'description': 'Number of recorded write operations',
            }
        }
    )

    read_bytes: Optional[int] = field(
        metadata={
            'metadata': {
                'description': 'Number of read bytes',
            }
        }
    )

    write_bytes: Optional[int] = field(
        metadata={
            'metadata': {
                'description': 'Number of written bytes',
            }
        }
    )

    read_time: Optional[float] = field(
        metadata={
            'metadata': {
                'description': 'Time spent reading, in seconds',
            }
        }
    )

    write_time: Optional[float] = field(
        metadata={
            'metadata': {
                'description': 'Time spent writing, in seconds',
            }
        }
    )

    busy_time: Optional[float] = field(
        metadata={
            'metadata': {
                'description': 'Total disk busy time, in seconds',
            }
        }
    )

    percent: float = percent_field()



from random import randint

from marshmallow import fields, missing
from marshmallow.schema import Schema
from marshmallow.validate import OneOf

from platypush.schemas import DateTime, Date, StrippedString

notification_types = ['follow', 'favourite', 'reblog', 'mention', 'poll', 'follow_request']
list_reply_policies = ['none', 'followed', 'list']


class MastodonSchema(Schema):
    pass


class MastodonAccountSchema(MastodonSchema):
    id = fields.String(
        dump_only=True,
        metadata=dict(
            example=''.join([f'{randint(1, 9)}' for _ in range(18)]),
        )
    )

    username = fields.String(
        metadata=dict(
            example='admin',
        )
    )

    url = fields.URL()
    avatar = fields.URL()
    header = fields.URL()
    followers_count = fields.Int(dump_only=True)
    following_count = fields.Int(dump_only=True)
    note = fields.String()
    display_name = StrippedString(
        metadata=dict(
            example='Name Surname',
        )
    )

    locked = fields.Boolean()
    bot = fields.Boolean()
    discoverable = fields.Boolean()
    group = fields.Boolean()
    created_at = DateTime(dump_only=True)
    last_status_at = DateTime(dump_only=True)


class MastodonFeaturedHashtagSchema(MastodonSchema):
    id = fields.Int(dump_only=True)
    name = fields.String()
    statuses_count = fields.Int(dump_only=True)
    last_status = DateTime(dump_only=True)


class MastodonHashtagHistorySchema(MastodonSchema):
    day = Date()
    uses = fields.Int()
    accounts = fields.Int()


class MastodonHashtagSchema(MastodonSchema):
    name = fields.String(metadata=dict(example='hashtag'))
    url = fields.URL()
    history = fields.Nested(
        MastodonHashtagHistorySchema, many=True, default=missing
    )


class MastodonMediaSchema(MastodonSchema):
    id = fields.String(dump_only=True)
    description = StrippedString()
    type = fields.String(dump_only=True, metadata={'example': 'image'})
    url = fields.URL(dump_only=True)
    preview_url = fields.URL(dump_only=True)
    remote_url = fields.URL(dump_only=True)
    preview_remote_url = fields.URL(dump_only=True)
    meta = fields.Dict()


class MastodonStatusSchema(MastodonSchema):
    id = fields.String(
        dump_only=True,
        metadata=dict(
            example=''.join([f'{randint(1, 9)}' for _ in range(18)]),
        )
    )

    in_reply_to_id = fields.String(
        dump_only=True,
        allow_none=True,
        metadata=dict(
            example=''.join([f'{randint(1, 9)}' for _ in range(18)]),
        )
    )

    in_reply_to_account_id = fields.String(
        dump_only=True,
        allow_none=True,
        metadata=dict(
            example=''.join([f'{randint(1, 9)}' for _ in range(18)]),
        )
    )

    url = fields.URL(dump_only=True)
    content = fields.String(allow_none=False)
    account = fields.Nested(MastodonAccountSchema, dump_only=True)
    attachments = fields.Nested(
        MastodonMediaSchema,
        many=True,
        dump_only=True,
        attribute='media_attachments',
    )
    hashtags = fields.Nested(
        MastodonHashtagSchema, many=True,
        attribute='tags', dump_only=True
    )

    replies_count = fields.Int(dump_only=True)
    reblogs_count = fields.Int(dump_only=True)
    favourites_count = fields.Int(dump_only=True)

    sensitive = fields.Boolean()
    favourited = fields.Boolean()
    reblogged = fields.Boolean()
    muted = fields.Boolean()
    bookmarked = fields.Boolean()
    pinned = fields.Boolean()

    created_at = DateTime(dump_only=True)


class MastodonSearchSchema(MastodonSchema):
    accounts = fields.Nested(MastodonAccountSchema, many=True)
    statuses = fields.Nested(MastodonStatusSchema, many=True)
    hashtags = fields.Nested(MastodonHashtagSchema, many=True)


class MastodonAccountCreationSchema(MastodonSchema):
    access_token = fields.String(dump_only=True)
    token_type = fields.String(dump_only=True, metadata={'example': 'Bearer'})
    scope = fields.String(dump_only=True, metadata={'example': 'read write follow push'})
    created_at = DateTime(dump_only=True)


class MastodonAccountListSchema(MastodonSchema):
    id = fields.Int(dump_only=True)
    title = StrippedString()


class MastodonFilterSchema(MastodonSchema):
    id = fields.Int(dump_only=True)
    phrase = StrippedString()
    whole_word = fields.Boolean()
    irreversible = fields.Boolean()
    expires_at = DateTime(allow_none=True)
    context = fields.List(
        fields.String(validate=OneOf(['home', 'notifications', 'public', 'thread'])),
        metadata={
            'example': 'Which context(s) this filter applies to. '
                       'Possible values: home, notifications, public, thread',
        }
    )


class MastodonConversationSchema(MastodonSchema):
    id = fields.Int(dump_only=True)
    unread = fields.Boolean()
    accounts = fields.Nested(MastodonAccountSchema, many=True)
    last_status = fields.Nested(MastodonStatusSchema)


class MastodonListSchema(MastodonSchema):
    id = fields.Int(dump_only=True)
    title = StrippedString()
    replies_policy = fields.String(validate=OneOf(list_reply_policies))


class MastodonMentionSchema(MastodonSchema):
    id = fields.Int(dump_only=True)
    username = StrippedString(metadata=dict(example='user'))
    url = fields.URL(metadata=dict(example='https://mastodon.social/@user'))


class MastodonNotificationSchema(MastodonSchema):
    id = fields.String(dump_only=True)
    type = fields.String(validate=OneOf(notification_types))
    account = fields.Nested(MastodonAccountSchema)
    status = fields.Nested(MastodonStatusSchema)
    mention = fields.Nested(MastodonMentionSchema)
    created_at = DateTime(dump_only=True)


class MastodonSubscriptionNotificationTypes(MastodonSchema):
    follow = fields.Boolean()
    reblog = fields.Boolean()
    mention = fields.Boolean()
    favourite = fields.Boolean()
    poll = fields.Boolean()



from typing import Optional, Union

from platypush.message.event import Event


class PushbulletEvent(Event):
    """
    Base PushBullet event.
    """

    def __init__(
        self,
        *args,
        notification_id: str,
        title: Optional[str] = None,
        body: Optional[Union[str, dict, list]] = None,
        url: Optional[str] = None,
        source_device: Optional[str] = None,
        source_user: Optional[str] = None,
        target_device: Optional[str] = None,
        icon: Optional[str] = None,
        created: float,
        modified: float,
        **kwargs
    ):
        """
        :param notification_id: Notification ID.
        :param title: Notification title.
        :param body: Notification body.
        :param url: Notification URL.
        :param source_device: Source device ID.
        :param source_user: Source user ID.
        :param target_device: Target device ID.
        :param icon: Notification icon.
        :param created: Notification creation timestamp.
        :param modified: Notification modification timestamp.
        :param kwargs: Additional attributes.
        """
        super().__init__(
            *args,
            notification_id=notification_id,
            title=title,
            body=body,
            url=url,
            source_device=source_device,
            source_user=source_user,
            target_device=target_device,
            icon=icon,
            created=created,
            modified=modified,
            **kwargs
        )


class PushbulletMessageEvent(PushbulletEvent):
    """
    Triggered when a new message is received.
    """

    def __init__(
        self,
        *args,
        sender_id: Optional[str] = None,
        sender_email: Optional[str] = None,
        sender_name: Optional[str] = None,
        receiver_id: Optional[str] = None,
        receiver_email: Optional[str] = None,
        receiver_name: Optional[str] = None,
        **kwargs
    ):
        super().__init__(
            *args,
            sender_id=sender_id,
            sender_email=sender_email,
            sender_name=sender_name,
            receiver_id=receiver_id,
            receiver_email=receiver_email,
            receiver_name=receiver_name,
            **kwargs
        )


class PushbulletNotificationEvent(PushbulletEvent):
    """
    Triggered when a notification is mirrored from another device.
    """

    def __init__(
        self,
        *args,
        title: str,
        body: str,
        dismissible: bool,
        application_name: Optional[str] = None,
        package_name: Optional[str] = None,
        actions: Optional[dict] = None,
        **kwargs
    ):
        """
        :param title: Mirror notification title.
        :param body: Mirror notification body.
        :param dismissible: True if the notification can be dismissed.
        :param application_name: Application name.
        :param package_name: Package name.
        :param actions: Actions associated to the notification. Example:

            .. code-block:: json

                [
                  {
                    "label": "previous",
                    "trigger_key": "com.termux.api_0_6107998_previous"
                  },
                  {
                    "label": "pause",
                    "trigger_key": "com.termux.api_0_6107998_pause"
                  },
                  {
                    "label": "play",
                    "trigger_key": "com.termux.api_0_6107998_play"
                  },
                  {
                    "label": "next",
                    "trigger_key": "com.termux.api_0_6107998_next"
                  }
                ]

        """
        super().__init__(
            *args,
            title=title,
            body=body,
            dismissible=dismissible,
            application_name=application_name,
            package_name=package_name,
            actions=actions,
            **kwargs
        )


class PushbulletDismissalEvent(PushbulletEvent):
    """
    Triggered when a notification is dismissed.
    """

    def __init__(self, *args, package_name: Optional[str] = None, **kwargs):
        super().__init__(*args, package_name=package_name, **kwargs)


class PushbulletLinkEvent(PushbulletMessageEvent):
    """
    Triggered when a push with a link is received.
    """


class PushbulletFileEvent(PushbulletMessageEvent):
    """
    Triggered when a push with a file is received.
    """

    def __init__(
        self,
        *args,
        file_name: str,
        file_type: str,
        file_url: str,
        image_width: Optional[int] = None,
        image_height: Optional[int] = None,
        image_url: Optional[str] = None,
        **kwargs
    ):
        super().__init__(
            *args,
            file_name=file_name,
            file_type=file_type,
            file_url=file_url,
            image_width=image_width,
            image_height=image_height,
            image_url=image_url,
            **kwargs
        )


# vim:sw=4:ts=4:et:



import logging
import threading
import select
import socket

from typing import Callable

logger = logging.getLogger(__name__)
should_run = {}


def handler(chan, local_port, host, port):
    key = local_port, host, port
    sock = socket.socket()
    try:
        sock.connect((host, port))
    except Exception as e:
        logger.warning('Forwarding request to {}:{} failed: {}'.format(host, port, e))
        return

    logger.info('Connected! Tunnel open {} -> {} -> {}'.format(
        chan.origin_addr, chan.getpeername(), (host, port)))

    while should_run.get(key):
        r, w, x = select.select([sock, chan], [], [])
        if sock in r:
            data = sock.recv(1024)
            if len(data) == 0:
                break
            chan.send(data)

        if chan in r:
            data = chan.recv(1024)
            if len(data) == 0:
                break
            sock.send(data)

    chan.close()
    sock.close()
    logger.info('Tunnel closed from {}'.format(chan.origin_addr))


def reverse_tunnel(server_port, remote_host, remote_port, transport, bind_addr='') -> Callable:
    def server():
        transport.request_port_forward(bind_addr, server_port)
        key = server_port, remote_host, remote_port
        should_run[key] = True

        while should_run.get(key):
            try:
                chan = transport.accept(1)
                assert chan is not None
            except Exception as e:
                logger.warning(e)
                continue

            thr = threading.Thread(
                target=handler, args=(chan, server_port, remote_host, remote_port))

            thr.setDaemon(True)
            thr.start()

    return server


def close_tunnel(server_port, remote_host, remote_port):
    key = server_port, remote_host, remote_port
    if key not in should_run:
        logger.info('No such active tunnel: {}:{}:{}'.format(server_port, remote_host, remote_port))

    del should_run[key]


# vim:sw=4:ts=4:et:



from contextlib import contextmanager
import os
import tempfile
from typing import Generator, Optional
from urllib.parse import urljoin

import requests

from platypush.plugins import action
from platypush.plugins.tts import TtsPlugin
from platypush.schemas.tts.mimic3 import Mimic3VoiceSchema


class TtsMimic3Plugin(TtsPlugin):
    r"""
    TTS plugin that uses the `Mimic3 webserver
    <https://github.com/MycroftAI/mimic3>`_ provided by `Mycroft
    <https://mycroft.ai/>`_ as a text-to-speech engine.

    The easiest way to deploy a Mimic3 instance is probably via Docker:

    .. code-block:: bash

        $ mkdir -p "$HOME/.local/share/mycroft/mimic3"
        $ chmod a+rwx "$HOME/.local/share/mycroft/mimic3"
        $ docker run --rm \
            -p 59125:59125 \
            -v "%h/.local/share/mycroft/mimic3:/home/mimic3/.local/share/mycroft/mimic3" \
            'mycroftai/mimic3'

    """

    def __init__(
        self,
        server_url: str,
        voice: str = 'en_US/vctk_low',
        **kwargs,
    ):
        """
        :param server_url: Base URL of the web server that runs the Mimic3 engine.
        :param voice: Default voice to be used (default: ``en_US/vctk_low``).
            You can get a full list of the voices available on the server
            through :meth:`.voices`.
        """
        super().__init__(**kwargs)
        self.server_url = server_url
        self.voice = voice
        self.player_args.update(
            {
                'channels': 1,
                'sample_rate': 22050,
                'dtype': 'int16',
            }
        )

    @staticmethod
    @contextmanager
    def _save_audio(
        text: str,
        server_url: str,
        voice: str,
        timeout: Optional[float] = None,
    ) -> Generator[str, None, None]:
        """
        Saves the raw audio stream from the Mimic3 server to an audio file for
        playback.

        :param text: Text to be spoken.
        :param server_url: Base URL of the Mimic3 server.
        :param voice: Voice to be used.
        :param timeout: Timeout for the audio stream retrieval.
        """

        rs = requests.post(
            urljoin(server_url, '/api/tts'),
            data=text,
            timeout=timeout,
            params={
                'voice': voice,
            },
        )

        rs.raise_for_status()
        tmp_file = tempfile.NamedTemporaryFile(suffix='.wav', delete=False)
        tmp_file.write(rs.content)
        yield tmp_file.name

        tmp_file.close()
        os.unlink(tmp_file.name)

    @action
    def say(
        self,
        text: str,
        *_,
        server_url: Optional[str] = None,
        voice: Optional[str] = None,
        **player_args,
    ):
        """
        Say some text.

        :param text: Text to say.
        :param server_url: Default ``server_url`` override.
        :param voice: Default ``voice`` override.
        :param player_args: Extends the additional arguments to be passed to
            :meth:`platypush.plugins.sound.SoundPlugin.play` (like volume,
            duration, channels etc.).
        """

        server_url = server_url or self.server_url
        voice = voice or self.voice

        with self._save_audio(text, server_url, voice) as audio_file:
            self._playback(audio_file, join=True, **player_args)

    @action
    def voices(self, server_url: Optional[str] = None):
        """
        List the voices available on the server.

        :param server_url: Default ``server_url`` override.
        :return: .. schema:: tts.mimic3.Mimic3VoiceSchema(many=True)
        """
        server_url = server_url or self.server_url
        rs = requests.get(urljoin(server_url, '/api/voices'), timeout=10)
        rs.raise_for_status()
        return Mimic3VoiceSchema().dump(rs.json(), many=True)


# vim:sw=4:ts=4:et:



from multiprocessing import Lock

from platypush.bus.redis import RedisBus
from platypush.context import get_bus
from platypush.config import Config
from platypush.message import Message
from platypush.message.request import Request
from platypush.utils import get_message_response

from .logger import logger


class BusWrapper:  # pylint: disable=too-few-public-methods
    """
    Lazy singleton wrapper for the bus object.
    """

    def __init__(self):
        self._redis_queue = None
        self._bus = None
        self._bus_lock = Lock()

    @property
    def bus(self) -> RedisBus:
        """
        Lazy getter/initializer for the bus object.
        """
        with self._bus_lock:
            if not self._bus:
                self._bus = get_bus()

        bus_: RedisBus = self._bus  # type: ignore
        return bus_

    def post(self, msg):
        """
        Send a message to the bus.

        :param msg: The message to send.
        """
        try:
            self.bus.post(msg)
        except Exception as e:
            logger().exception(e)


_bus = BusWrapper()


def bus():
    """
    Lazy getter/initializer for the bus object.
    """
    return _bus.bus


def send_message(msg, wait_for_response=True):
    """
    Send a message to the bus.

    :param msg: The message to send.
    :param wait_for_response: If ``True``, wait for the response to be received
        before returning, otherwise return immediately.
    """
    msg = Message.build(msg)
    if msg is None:
        return None

    if isinstance(msg, Request):
        msg.origin = 'http'

    if Config.get('token'):
        msg.token = Config.get('token')

    bus().post(msg)

    if isinstance(msg, Request) and wait_for_response:
        response = get_message_response(msg)
        logger().debug('Processing response on the HTTP backend: %s', response)

        return response

    return None


def send_request(action, wait_for_response=True, **kwargs):
    """
    Send a request to the bus.

    :param action: The action to send.
    :param wait_for_response: If ``True``, wait for the response to be received
        before returning, otherwise return immediately.
    :param kwargs: Additional arguments to pass to the action.
    """
    msg = {'type': 'request', 'action': action}

    if kwargs:
        msg['args'] = kwargs

    rs = send_message(msg, wait_for_response=wait_for_response)
    assert rs, 'Got an empty response from the server'
    if rs:
        assert not rs.errors, '\n'.join(rs.errors)

    return rs.output



import datetime
from typing import Optional

import requests
from dateutil.tz import gettz

from platypush.plugins import Plugin, action
from platypush.plugins.calendar import CalendarInterface
from platypush.utils import utcnow


class CalendarIcalPlugin(Plugin, CalendarInterface):
    """
    iCal calendars plugin. Interact with remote calendars in iCal format.
    """

    def __init__(self, url, *args, **kwargs):
        """
        :param url: iCal URL to parse
        :type url: str
        """
        super().__init__(*args, **kwargs)
        self.url = url

    @staticmethod
    def _convert_timestamp(event: dict, attribute: str) -> Optional[str]:
        t = event.get(attribute)
        if not t:
            return None

        if isinstance(t.dt, datetime.date) and not isinstance(t.dt, datetime.datetime):
            return datetime.datetime(
                t.dt.year, t.dt.month, t.dt.day, tzinfo=gettz()
            ).isoformat()

        return (
            datetime.datetime.fromtimestamp(t.dt.timestamp())
            .replace(tzinfo=t.dt.tzinfo or gettz())
            .astimezone(datetime.timezone.utc)
            .isoformat()
        )

    @classmethod
    def _translate_event(cls, event):
        return {
            'id': str(event.get('uid')) if event.get('uid') else None,
            'kind': 'calendar#event',
            'summary': str(event.get('summary')) if event.get('summary') else None,
            'description': (
                str(event.get('description')) if event.get('description') else None
            ),
            'status': str(event.get('status')).lower() if event.get('status') else None,
            'responseStatus': (
                str(event.get('partstat')).lower() if event.get('partstat') else None
            ),
            'location': str(event.get('location')) if event.get('location') else None,
            'htmlLink': str(event.get('url')) if event.get('url') else None,
            'organizer': (
                {
                    'email': str(event.get('organizer')).replace('MAILTO:', ''),
                    'displayName': event.get('organizer').params.get('cn'),
                }
                if event.get('organizer')
                else None
            ),
            'created': cls._convert_timestamp(event, 'created'),
            'updated': cls._convert_timestamp(event, 'last-modified'),
            'start': {
                'dateTime': cls._convert_timestamp(event, 'dtstart'),
                'timeZone': 'UTC',
            },
            'end': {
                'dateTime': cls._convert_timestamp(event, 'dtend'),
                'timeZone': 'UTC',
            },
        }

    @action
    def get_upcoming_events(self, *_, only_participating=True, **__):
        """
        Get the upcoming events. See
        :meth:`platypush.plugins.calendar.CalendarPlugin.get_upcoming_events`.
        """

        from icalendar import Calendar

        events = []
        response = requests.get(self.url, timeout=20)
        assert (
            response.ok
        ), f"HTTP error while getting events from {self.url}: {response.text}"

        calendar = Calendar.from_ical(response.text)
        for event in calendar.walk():
            if event.name != 'VEVENT':
                continue  # Not an event

            event = self._translate_event(event)

            if (
                event['status'] != 'cancelled'
                and event['end'].get('dateTime')
                and datetime.datetime.fromisoformat(event['end']['dateTime'])
                >= utcnow()
                and (
                    (
                        only_participating
                        and event.get('responseStatus')
                        in [None, 'accepted', 'tentative']
                    )
                    or not only_participating
                )
            ):
                events.append(event)

        return events


# vim:sw=4:ts=4:et:



import logging
from threading import Event
from typing import Collection, Optional

from platypush.utils import utcnow

from ._base import (
    Entity,
    EntityKey,
    EntitySavedCallback,
    get_entities_registry,
    init_entities_db,
)
from ._engine import EntitiesEngine
from .managers import (
    EntityManager,
    get_plugin_entity_registry,
    register_entity_manager,
)
from .managers.lights import LightEntityManager
from .managers.sensors import SensorEntityManager
from .managers.switches import (
    DimmerEntityManager,
    EnumSwitchEntityManager,
    SwitchEntityManager,
)

_engine: Optional[EntitiesEngine] = None
logger = logging.getLogger(__name__)


def init_entities_engine() -> EntitiesEngine:
    """
    Initialize and start the entities engine.
    """
    global _engine  # pylint: disable=global-statement
    init_entities_db()
    _engine = EntitiesEngine()
    _engine.start()
    return _engine


def get_entities_engine(timeout: Optional[float] = None) -> EntitiesEngine:
    """
    Return the running entities engine.

    :param timeout: Timeout in seconds (default: None).
    """
    time_start = utcnow().timestamp()
    while not timeout or (utcnow().timestamp() - time_start < timeout):
        if _engine:
            break

        Event().wait(1)

    if not _engine:
        raise TimeoutError('The entities engine has not been initialized')
    return _engine


def publish_entities(
    entities: Collection[Entity], callback: Optional[EntitySavedCallback] = None
) -> None:
    """
    Publish a collection of entities to be processed by the engine.

    The engine will:

        - Normalize and merge the provided entities.
        - Trigger ``EntityUpdateEvent`` events.
        - Persist the new state to the local database.

    :param entities: Entities to be published.
    """
    if not _engine:
        logger.debug('No entities engine registered')
        return

    _engine.post(*entities, callback=callback)


__all__ = (
    'DimmerEntityManager',
    'EntitiesEngine',
    'Entity',
    'EntityKey',
    'EntityManager',
    'EntitySavedCallback',
    'EnumSwitchEntityManager',
    'LightEntityManager',
    'SensorEntityManager',
    'SwitchEntityManager',
    'get_entities_registry',
    'get_plugin_entity_registry',
    'init_entities_engine',
    'publish_entities',
    'register_entity_manager',
)



import re
from typing import Optional
from urllib.parse import urlparse

from dns.exception import DNSException
from dns.resolver import resolve

from ._account import AccountConfig
from ._model import ServerConfig

_email_regex = re.compile(r'[^@]+@[^@]+\.[^@]+')
_from_header_regex = re.compile(r'^(?:"?([^"]+)"?\s+)?<?([^>]+)>?$')
_ipv4_regex = re.compile(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$')
_ipv6_regex = re.compile(r'^[0-9a-fA-F:]+$')


def infer_mail_domain(account: AccountConfig, server: ServerConfig) -> str:
    """
    Infers the mail domain from the account and server configuration.
    """

    if server.domain:
        return server.domain

    if account.username and _email_regex.match(account.username):
        return account.username.split('@', 1)[1]

    if server.server:
        if _ipv4_regex.match(server.server) or _ipv6_regex.match(server.server):
            return server.server

        host = urlparse(server.server).hostname
        assert host, f'Could not parse hostname from server URL: {server.server}'
        host_tokens = host.split('.')

        while host_tokens:
            try:
                resolve(host, 'MX')
                return host
            except DNSException:
                host_tokens.pop(0)
                host = '.'.join(host_tokens)

        raise AssertionError(f'Could not resolve MX record for {host}')

    raise AssertionError('Could not infer mail domain from configuration.')


def infer_mail_address(account: AccountConfig, server: ServerConfig) -> str:
    if account.username and _email_regex.match(account.username):
        return account.username

    return f'{account.username}@{infer_mail_domain(account, server)}'


def normalize_from_header(
    from_: Optional[str], account: AccountConfig, server: ServerConfig
) -> str:
    """
    Normalizes the value of the "From" header.
    """

    if not from_:
        from_ = account.display_name or account.username

    if _email_regex.match(from_):
        return from_

    m = _from_header_regex.match(from_)
    if m and _email_regex.match(m.group(2)):
        return f'{m.group(1)} <{m.group(2)}>'

    return f'{from_} <{infer_mail_address(account, server)}>'



from dataclasses import dataclass
import datetime

from typing import Optional, List, Union


@dataclass
class TrelloList:
    """
    Represents a Trello list.
    """

    id: str
    name: str
    closed: bool
    subscribed: bool


@dataclass
class TrelloBoard:
    """
    Represents a Trello board.
    """

    id: str
    name: str
    url: str
    closed: bool
    lists: Optional[List[TrelloList]] = None
    description: Optional[str] = None
    date_last_activity: Optional[datetime.datetime] = None


@dataclass
class TrelloLabel:
    """
    Represents a Trello label.
    """

    id: str
    name: str
    color: Optional[str] = None


@dataclass
class TrelloUser:
    """
    Represents a Trello user.
    """

    id: str
    username: str
    fullname: str
    initials: Optional[str] = None
    avatar_url: Optional[str] = None


@dataclass
class TrelloComment:
    """
    Represents a Trello comment.
    """

    id: str
    text: str
    type: str
    creator: TrelloUser
    date: Union[str, datetime.datetime]


@dataclass
class TrelloPreview:
    """
    Represents a Trello preview.
    """

    id: str
    url: str
    scaled: bool
    size: int
    height: int
    width: int


@dataclass
class TrelloAttachment:
    """
    Represents a Trello attachment.
    """

    id: str
    size: int
    date: str
    edge_color: str
    member_id: str
    is_upload: bool
    name: str
    previews: List[TrelloPreview]
    url: str
    mime_type: Optional[str] = None


@dataclass
class TrelloChecklistItem:
    """
    Represents a Trello checklist item.
    """

    id: str
    name: str
    checked: bool


@dataclass
class TrelloChecklist:
    """
    Represents a Trello checklist.
    """

    id: str
    name: str
    items: List[TrelloChecklistItem]


@dataclass
class TrelloCard:
    """
    Represents a Trello card.
    """

    id: str
    name: str
    url: str
    closed: bool
    board: TrelloBoard
    is_due_complete: bool
    description: Optional[str] = None
    list: Optional[TrelloList] = None
    comments: Optional[List[TrelloComment]] = None
    labels: Optional[List[TrelloLabel]] = None
    attachments: Optional[List[TrelloAttachment]] = None
    checklists: Optional[List[TrelloChecklist]] = None
    due_date: Optional[Union[datetime.datetime, str]] = None
    latest_card_move_date: Optional[Union[datetime.datetime, str]] = None
    date_last_activity: Optional[Union[datetime.datetime, str]] = None


@dataclass
class TrelloMember:
    """
    Represents a Trello member.
    """

    id: str
    fullname: str
    bio: Optional[str]
    url: Optional[str]
    username: Optional[str]
    initials: Optional[str]
    member_type: Optional[str] = None


# vim:sw=4:ts=4:et:



from marshmallow import Schema, fields


class Mimic3Schema(Schema):
    pass


class Mimic3VoiceSchema(Mimic3Schema):
    key = fields.String(
        required=True,
        dump_only=True,
        metadata={
            'description': 'Unique voice ID',
            'example': 'en_UK/apope_low',
        },
    )

    language = fields.String(
        required=True,
        dump_only=True,
        metadata={
            'example': 'en_UK',
        },
    )

    language_english = fields.String(
        metadata={
            'description': 'Name of the language (in English)',
        }
    )

    language_native = fields.String(
        metadata={
            'description': 'Name of the language (in the native language)',
        }
    )

    name = fields.String(
        metadata={
            'example': 'apope_low',
        }
    )

    sample_text = fields.String(
        metadata={
            'example': 'Some text',
        }
    )

    description = fields.String()
    aliases = fields.List(fields.String)



from typing import Type

from platypush.common.gstreamer import Pipeline
from platypush.context import get_bus
from platypush.message.event.media import MediaEvent, MediaPlayEvent, MediaPauseEvent, MediaStopEvent, \
    NewPlayingMediaEvent, MediaMuteChangedEvent, MediaSeekEvent


class MediaPipeline(Pipeline):
    def __init__(self, resource: str):
        super().__init__()
        self.resource = resource
        self.add_source('playbin', uri=resource)

    @staticmethod
    def post_event(evt_class: Type[MediaEvent], **kwargs):
        kwargs['player'] = 'local'
        kwargs['plugin'] = 'media.gstreamer'
        evt = evt_class(**kwargs)
        get_bus().post(evt)

    def play(self):
        # noinspection PyUnresolvedReferences,PyPackageRequirements
        from gi.repository import Gst
        is_first_play = self.get_state() == Gst.State.NULL

        super().play()
        if is_first_play:
            self.post_event(NewPlayingMediaEvent, resource=self.resource)
        self.post_event(MediaPlayEvent, resource=self.resource)

    def pause(self):
        # noinspection PyUnresolvedReferences,PyPackageRequirements
        from gi.repository import Gst
        super().pause()
        self.post_event(MediaPauseEvent if self.get_state() == Gst.State.PAUSED else MediaPlayEvent)

    def stop(self):
        super().stop()
        self.post_event(MediaStopEvent)

    def mute(self):
        super().mute()
        self.post_event(MediaMuteChangedEvent, mute=self.is_muted())

    def unmute(self):
        super().unmute()
        self.post_event(MediaMuteChangedEvent, mute=self.is_muted())

    def seek(self, position: float):
        super().seek(position)
        self.post_event(MediaSeekEvent, position=self.get_position())


# vim:sw=4:ts=4:et:



import os
import tempfile
from contextlib import contextmanager
from multiprocessing import Process
from typing import Generator, Optional

import requests

from platypush.context import get_plugin
from platypush.plugins import action
from platypush.plugins.openai import OpenaiPlugin
from platypush.plugins.tts import TtsPlugin


class TtsOpenaiPlugin(TtsPlugin):
    r"""
    This plugin provides an interface to the `OpenAI text-to-speech API
    <https://platform.openai.com/docs/guides/text-to-speech>`_.

    It requires the :class:`platypush.plugins.openai.OpenaiPlugin` plugin to be
    configured.
    """

    _BUFSIZE = 1024

    def __init__(
        self,
        model: str = 'tts-1',
        voice: str = 'nova',
        timeout: float = 10,
        **kwargs,
    ):
        """
        :param model: Model to be used for the text-to-speech conversion.
            See the `OpenAI API models documentation
            <https://platform.openai.com/docs/models/tts>`_ for the list of
            available models (default: ``tts-1``).
        :param voice: Default voice to be used. See the `OpenAI API
            voices documentation
            <https://platform.openai.com/docs/guides/text-to-speech/voice-options>`_
            for the list of available voices (default: ``nova``).
        :param timeout: Default timeout for the API requests (default: 10s).
        """
        super().__init__(**kwargs)
        openai = get_plugin('openai')
        assert openai, 'openai plugin not configured'

        self.openai: OpenaiPlugin = openai
        self.model = model
        self.voice = voice
        self.timeout = timeout
        self._audio_proc: Optional[Process] = None

    def _process_response(
        self,
        response: requests.Response,
        audio_file: str,
    ) -> Process:
        def proc_fn():
            try:
                with open(audio_file, 'wb') as file:
                    for chunk in response.iter_content(chunk_size=self._BUFSIZE):
                        if chunk:
                            file.write(chunk)
                            file.flush()
            except KeyboardInterrupt:
                pass

        self._audio_proc = Process(target=proc_fn, name='openai-tts-response-processor')
        self._audio_proc.start()
        return self._audio_proc

    def _make_request(
        self,
        text: str,
        model: Optional[str] = None,
        voice: Optional[str] = None,
    ) -> requests.Response:
        rs = requests.post(
            "https://api.openai.com/v1/audio/speech",
            timeout=self.timeout,
            stream=True,
            headers={
                "Authorization": f"Bearer {self.openai._api_key}",  # pylint: disable=protected-access
                "Content-Type": "application/json",
            },
            json={
                "model": model or self.model,
                "voice": voice or self.voice,
                "input": text,
            },
        )

        rs.raise_for_status()
        return rs

    @contextmanager
    def _audio_fifo(self) -> Generator[str, None, None]:
        fifo_dir = tempfile.mkdtemp()
        fifo_path = os.path.join(fifo_dir, 'platypush-tts-openai-fifo')
        os.mkfifo(fifo_path)
        yield fifo_path

        os.unlink(fifo_path)
        os.rmdir(fifo_dir)

    @action
    def say(
        self,
        text: str,
        *_,
        model: Optional[str] = None,
        voice: Optional[str] = None,
        **player_args,
    ):
        """
        Say some text.

        :param text: Text to say.
        :param model: Default ``model`` override.
        :param voice: Default ``voice`` override.
        :param player_args: Extends the additional arguments to be passed to
            :meth:`platypush.plugins.sound.SoundPlugin.play` (like volume,
            duration, channels etc.).
        """
        response_processor: Optional[Process] = None

        try:
            response = self._make_request(text, model=model, voice=voice)

            with self._audio_fifo() as audio_file:
                response_processor = self._process_response(
                    response=response, audio_file=audio_file
                )
                self._playback(audio_file, **player_args)
                response_processor.join()
                response_processor = None
        finally:
            if response_processor:
                response_processor.terminate()

    @action
    def stop(self):
        super().stop()
        if self._audio_proc and self._audio_proc.is_alive():
            self._audio_proc.terminate()
            self._audio_proc.join()


# vim:sw=4:ts=4:et:



import asyncio as aio
import logging
import os
import re

from multiprocessing import Event, Process, Queue
from typing import Callable, Coroutine, Optional, Set, Type, Union

from telegram import Message, Update
from telegram.ext import (
    Application,
    ApplicationBuilder,
    ContextTypes,
    MessageHandler,
    filters,
)

from platypush.context import get_bus
from platypush.message.event.telegram import (
    MessageEvent,
    CommandMessageEvent,
    TextMessageEvent,
    PhotoMessageEvent,
    VideoMessageEvent,
    ContactMessageEvent,
    DocumentMessageEvent,
    LocationMessageEvent,
    GroupChatCreatedEvent,
)

from ._bridge import CommandBridge
from ._model import END_OF_SERVICE, Resource
from ._utils import dump_msg, dump_user


class TelegramService(Process):
    """
    Background service to handle Telegram messages and events.
    """

    def __init__(
        self,
        api_token: str,
        cmd_queue: Queue,
        result_queue: Queue,
        authorized_chat_ids: Set[Union[str, int]],
        *_,
        **__,
    ):
        super().__init__(name="telegram-service")
        self.logger = logging.getLogger("platypush:telegram")
        self.authorized_chat_ids = set(authorized_chat_ids or [])
        self._cmd_queue = cmd_queue
        self._result_queue = result_queue
        self._loop = aio.new_event_loop()
        self._app = self._build_app(api_token)
        self._service_running = Event()
        self._service_stopped = Event()
        self._cmd_bridge = CommandBridge(self)

    def _build_app(self, api_token: str) -> Application:
        app = ApplicationBuilder().token(api_token).build()
        app.add_handler(MessageHandler(filters.ChatType.GROUPS, self._group_hook()))

        app.add_handler(
            MessageHandler(
                filters.TEXT & (~filters.COMMAND), self._msg_hook(TextMessageEvent)
            )
        )

        app.add_handler(
            MessageHandler(filters.PHOTO, self._msg_hook(PhotoMessageEvent))
        )

        app.add_handler(
            MessageHandler(filters.VIDEO, self._msg_hook(VideoMessageEvent))
        )

        app.add_handler(
            MessageHandler(filters.CONTACT, self._msg_hook(ContactMessageEvent))
        )

        app.add_handler(
            MessageHandler(filters.LOCATION, self._msg_hook(LocationMessageEvent))
        )

        app.add_handler(
            MessageHandler(filters.Document.ALL, self._msg_hook(DocumentMessageEvent))
        )

        app.add_handler(MessageHandler(filters.COMMAND, self._command_hook()))
        return app

    async def _authorize(self, msg: Message, context: ContextTypes.DEFAULT_TYPE):
        if not self.authorized_chat_ids:
            return

        if msg.chat.type == 'private' and msg.chat.id not in self.authorized_chat_ids:
            self.logger.info(
                'Received message from unauthorized chat_id %s', msg.chat.id
            )

            await context.bot.send_message(
                chat_id=msg.chat.id,
                text='You are not allowed to send messages to this bot',
            )

            raise PermissionError

    def _msg_hook(self, cls: Type[MessageEvent]):
        async def hook(update: Update, context: ContextTypes.DEFAULT_TYPE):
            msg = update.effective_message
            if not msg:
                return

            try:
                await self._authorize(msg, context)
                self.bus.post(
                    cls(
                        chat_id=(
                            update.effective_chat.id if update.effective_chat else None
                        ),
                        message=dump_msg(msg) if msg else None,
                        user=(
                            dump_user(update.effective_user)
                            if update.effective_user
                            else None
                        ),
                    )
                )
            except PermissionError:
                pass

        return hook

    def _group_hook(self):
        async def hook(update: Update, context):
            msg = update.effective_message
            if not msg:
                return

            if msg.group_chat_created:
                self.bus.post(
                    GroupChatCreatedEvent(
                        chat_id=(
                            update.effective_chat.id if update.effective_chat else None
                        ),
                        message=dump_msg(msg),
                        user=(
                            dump_user(update.effective_user)
                            if update.effective_user
                            else None
                        ),
                    )
                )
            elif msg.photo:
                await self._msg_hook(PhotoMessageEvent)(update, context)
            elif msg.video:
                await self._msg_hook(VideoMessageEvent)(update, context)
            elif msg.contact:
                await self._msg_hook(ContactMessageEvent)(update, context)
            elif msg.location:
                await self._msg_hook(LocationMessageEvent)(update, context)
            elif msg.document:
                await self._msg_hook(DocumentMessageEvent)(update, context)
            elif msg.text:
                if msg.text.startswith('/'):
                    await self._command_hook()(update, context)
                else:
                    await self._msg_hook(TextMessageEvent)(update, context)

        return hook

    def _command_hook(self):
        async def hook(update: Update, context: ContextTypes.DEFAULT_TYPE):
            msg = update.effective_message
            if not (msg and msg.text):
                return

            m = re.match(r'\s*/([0-9a-zA-Z_-]+)\s*(.*)', msg.text)
            if not m:
                self.logger.warning('Invalid command: %s', msg.text)
                return

            cmd = m.group(1).lower()
            args = [arg for arg in re.split(r'\s+', m.group(2)) if len(arg)]

            try:
                await self._authorize(msg, context)
                self.bus.post(
                    CommandMessageEvent(
                        chat_id=(
                            update.effective_chat.id if update.effective_chat else None
                        ),
                        command=cmd,
                        cmdargs=args,
                        message=dump_msg(msg),
                        user=(
                            dump_user(update.effective_user)
                            if update.effective_user
                            else None
                        ),
                    )
                )
            except PermissionError:
                pass

        return hook

    def _exec(
        self,
        method: Callable[..., Coroutine],
        *args,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        fut = aio.run_coroutine_threadsafe(method(*args, **kwargs), self._loop)
        return fut.result(timeout=timeout)

    def exec(
        self,
        cmd: str,
        *args,
        timeout: Optional[float] = None,
        resource: Optional[Resource] = None,
        resource_attr: Optional[str] = None,
        **kwargs,
    ):
        method = getattr(self._app.bot, cmd, None)
        assert method, f"Method {cmd} not found"

        if resource:
            assert resource_attr, f"Resource attribute not specified for command {cmd}"
            with resource as file:
                kwargs[resource_attr] = file
                return self._exec(
                    method,
                    *args,
                    timeout=timeout,
                    **kwargs,
                )

        return self._exec(method, *args, timeout=timeout, **kwargs)

    def _run(self):
        self._app.run_polling()

    def run(self):
        super().run()
        self._service_running.set()
        self._service_stopped.clear()
        aio.set_event_loop(self._loop)
        self._cmd_bridge.start()

        try:
            self._run()
        except Exception as e:
            self.logger.error("Telegram polling error: %s", e, exc_info=True)
        finally:
            self._service_running.clear()
            self._service_stopped.set()
            self._cmd_queue.put_nowait(END_OF_SERVICE)

    def stop(self):
        self._cmd_queue.put_nowait(END_OF_SERVICE)
        self._app.stop_running()

        if self.is_alive() and self.pid != os.getpid():
            self.terminate()
            self.join(timeout=5)

            if self.is_alive():
                self.kill()

    @property
    def bus(self):
        return get_bus()

    @property
    def cmd_queue(self):
        return self._cmd_queue

    @property
    def result_queue(self):
        return self._result_queue

    @property
    def stop_event(self):
        return self._service_stopped

    def is_running(self):
        return self._service_running.is_set()

    @property
    def run_event(self):
        return self._service_running


# vim:sw=4:ts=4:et:



import inspect
import os
import pathlib
import random
import threading
import time
from typing import Dict, Iterable, Optional, Union

import requests

from platypush.context import get_bus
from platypush.plugins import Plugin, action
from platypush.message.event.torrent import (
    TorrentDownloadStartEvent,
    TorrentDownloadedMetadataEvent,
    TorrentStateChangeEvent,
    TorrentDownloadProgressEvent,
    TorrentDownloadCompletedEvent,
    TorrentDownloadStopEvent,
    TorrentPausedEvent,
    TorrentResumedEvent,
    TorrentQueuedEvent,
)
from platypush.utils import get_default_downloads_dir

from . import _search as search_module
from ._search import TorrentSearchProvider


class TorrentPlugin(Plugin):
    """
    Plugin to search and download torrents.

    Search
    ------

    You can search for torrents using the :meth:`search` method. The method will
    use the search providers configured in the ``search_providers`` attribute of
    the plugin configuration. Currently supported search providers:

        * ``popcorntime``:
            :class:`platypush.plugins.torrent._search.PopcornTimeSearchProvider`
        * ``torrents.csv``:
            :class:`platypush.plugins.torrent._search.TorrentsCsvSearchProvider`

    ``torrents.csv`` will be enabled by default unless you explicitly disable
    it. ``torrents.csv`` also supports both:

        * A remote API via the ``api_url`` attribute (default:
           `https://torrents-csv.com/service``). You can also run your own API
           server by following the instructions at `heretic/torrents-csv-server
           <https://git.torrents-csv.com/heretic/torrents-csv-server>`_.

        * A local checkout of the ``torrents.csv`` file. Clone the
          `heretic/torrents-csv-data
          <https://git.torrents-csv.com/heretic/torrents-csv-data>`_ and provide
          the path to the ``torrents.csv`` file in the ``csv_file`` attribute.

        * A local checkout of the ``torrents.db`` file built from the
          ``torrents.csv`` file. Follow the instructions at
          `heretic/torrents-csv-data
          <https://git.torrents-csv.com/heretic/torrents-csv-data>`_ on how to
          build the ``torrents.db`` file from the ``torrents.csv`` file.

    If you opt for a local checkout of the ``torrents.csv`` file, then
    Platypush will build the SQLite database from the CSV file for you - no need
    to use external services. This however means that the first search will be
    slower as the database is being built. Subsequent searches will be faster,
    unless you modify the CSV file - in this case, an updated database will be
    built from the latest CSV file.

    You can also specify the ``download_csv`` property in the configuration. In
    this case, Platypush will automatically download the latest ``torrents.csv``
    file locally and build the SQLite database from it. On startup, Platypush
    will check if either the local or remote CSV file has been updated, and
    rebuild the database if necessary.

    ``popcorntime`` will be disabled by default unless you explicitly enable it.
    That's because, at the time of writing (June 2024), there are no publicly
    available PopcornTime API servers. You can run your own PopcornTime API
    server by following the instructions at `popcorn-time-ru/popcorn-ru
    <https://github.com/popcorn-time-ru/popcorn-ru>`_.

    Configuration example:

    .. code-block:: yaml

        torrent:
          # ...

          search_providers:
            torrents.csv:
              # Default: True
              # enabled: true
              # Base URL of the torrents.csv API.
              api_url: https://torrents-csv.com/service

              # Alternatively, you can also use a local checkout of the
              # torrents.csv file.
              # csv_file: /path/to/torrents.csv

              # Or a manually built SQLite database from the torrents.csv file.
              # db_file: /path/to/torrents.db

              # Or automatically download the latest torrents.csv file.
              # download_csv: true
            popcorn_time:
              # Default: false
              # enabled: false
              # Required: PopcornTime API base URL.
              api_url: https://popcorntime.app

    """

    _http_timeout = 20

    # Wait time in seconds between two torrent transfer checks
    _MONITOR_CHECK_INTERVAL = 3

    default_torrent_ports = (6881, 6891)
    torrent_state = {}
    transfers = {}

    def __init__(
        self,
        download_dir: Optional[str] = None,
        torrent_ports: Iterable[int] = default_torrent_ports,
        search_providers: Optional[
            Union[Dict[str, dict], Iterable[TorrentSearchProvider]]
        ] = None,
        **kwargs,
    ):
        """
        :param download_dir: Directory where the videos/torrents will be
            downloaded (default: ``~/Downloads``).
        :param torrent_ports: Torrent ports to listen on (default: 6881 and 6891)
        :param search_providers: List of search providers to use.
        """
        super().__init__(**kwargs)

        self.torrent_ports = torrent_ports
        self.download_dir = os.path.abspath(
            os.path.expanduser(download_dir or get_default_downloads_dir())
        )

        self._search_providers = self._load_search_providers(search_providers)
        self.logger.info(
            'Loaded search providers: %s',
            [provider.provider_name() for provider in self._search_providers],
        )

        self._sessions = {}
        self._lt_session = None
        pathlib.Path(self.download_dir).mkdir(parents=True, exist_ok=True)

    def _load_search_providers(
        self,
        search_providers: Optional[
            Union[Dict[str, dict], Iterable[TorrentSearchProvider]]
        ],
    ) -> Iterable[TorrentSearchProvider]:
        provider_classes = {
            cls.provider_name(): cls
            for _, cls in inspect.getmembers(search_module, inspect.isclass)
            if issubclass(cls, TorrentSearchProvider) and cls != TorrentSearchProvider
        }

        if not search_providers:
            return [
                provider()
                for provider in provider_classes.values()
                if provider.default_enabled()
            ]

        parsed_providers = []
        if isinstance(search_providers, dict):
            providers_dict = {}

            # Configure the search providers explicitly passed in the configuration
            for provider_name, provider_config in search_providers.items():
                if provider_name not in provider_classes:
                    self.logger.warning(
                        'Unsupported search provider %s. Supported providers: %s',
                        provider_name,
                        list(provider_classes.keys()),
                    )
                    continue

                provider_class = provider_classes[provider_name]
                providers_dict[provider_name] = provider_class(
                    **{'enabled': True, **provider_config}
                )

            # Enable the search providers that have `default_enabled` set to True
            # and that are not explicitly disabled
            for provider_name, provider in provider_classes.items():
                if provider.default_enabled() and provider_name not in search_providers:
                    providers_dict[provider_name] = provider()

            parsed_providers = list(providers_dict.values())
        else:
            parsed_providers = search_providers

        assert all(
            isinstance(provider, TorrentSearchProvider) for provider in parsed_providers
        ), 'All search providers must be instances of TorrentSearchProvider'

        return parsed_providers

    @action
    def search(
        self,
        query: str,
        *args,
        providers: Optional[Union[str, Iterable[str]]] = None,
        limit: int = 25,
        page: int = 1,
        **filters,
    ):
        """
        Perform a search of video torrents.

        :param query: Query string, video name or partial name
        :param providers: Override the default search providers by specifying
            the provider names to use for this search.
        :param filters: Additional filters to apply to the search, depending on
            what the configured search providers support. For example,
            ``category`` and ``language`` are supported by the PopcornTime.
        :param limit: Maximum number of results to return (default: 25).
        :param page: Page number (default: 1).
        :return: .. schema:: torrent.TorrentResultSchema(many=True)
        """
        results = []

        def worker(provider: TorrentSearchProvider):
            self.logger.debug(
                'Searching torrents on provider %s, query: %s',
                provider.provider_name(),
                query,
            )
            results.extend(
                provider.search(query, *args, limit=limit, page=page, **filters)
            )

        if providers:
            providers = [providers] if isinstance(providers, str) else providers
            search_providers = [
                provider
                for provider in self._search_providers
                if provider.provider_name() in providers
            ]
        else:
            search_providers = self._search_providers

        if not search_providers:
            self.logger.warning('No search providers enabled')
            return []

        workers = [
            threading.Thread(target=worker, kwargs={'provider': provider})
            for provider in search_providers
        ]

        for wrk in workers:
            wrk.start()
        for wrk in workers:
            wrk.join()

        return [result.to_dict() for result in results]

    def _get_torrent_info(self, torrent, download_dir):
        import libtorrent as lt

        torrent_file = None
        magnet = None
        info = {}
        file_info = {}

        if torrent.startswith('magnet:?'):
            magnet = torrent
            magnet_info = lt.parse_magnet_uri(magnet)
            if isinstance(magnet_info, dict):
                info = {
                    'name': magnet_info.get('name'),
                    'url': magnet,
                    'magnet': magnet,
                    'trackers': magnet_info.get('trackers', []),
                    'save_path': download_dir,
                }
            else:
                info = {
                    'name': magnet_info.name,
                    'url': magnet,
                    'magnet': magnet,
                    'trackers': magnet_info.trackers,
                    'save_path': download_dir,
                }
        elif torrent.startswith('http://') or torrent.startswith('https://'):
            response = requests.get(
                torrent, timeout=self._http_timeout, allow_redirects=True
            )
            torrent_file = os.path.join(download_dir, self._generate_rand_filename())

            with open(torrent_file, 'wb') as f:
                f.write(response.content)
        else:
            torrent_file = os.path.abspath(os.path.expanduser(torrent))
            if not os.path.isfile(torrent_file):
                raise RuntimeError(f'{torrent_file} is not a valid torrent file')

        if torrent_file:
            file_info = lt.torrent_info(torrent_file)
            info = {
                'name': file_info.name(),
                'url': torrent,
                'trackers': [t.url for t in list(file_info.trackers())],
                'save_path': download_dir,
            }

        return info, file_info, torrent_file, magnet

    def _fire_event(self, event, event_hndl):
        bus = get_bus()
        bus.post(event)

        try:
            if event_hndl:
                event_hndl(event)
        except Exception as e:
            self.logger.warning('Exception in torrent event handler: %s', e)
            self.logger.exception(e)

    def _torrent_monitor(self, torrent, transfer, download_dir, event_hndl, is_media):
        def thread():
            files = []
            last_status = None
            download_started = False
            metadata_downloaded = False

            while not transfer.is_finished():
                if torrent not in self.transfers:
                    self.logger.info('Torrent %s has been stopped and removed', torrent)
                    self._fire_event(TorrentDownloadStopEvent(url=torrent), event_hndl)
                    break

                status = transfer.status()
                torrent_file = transfer.torrent_file()

                if torrent_file:
                    self.torrent_state[torrent]['size'] = torrent_file.total_size()
                    files = [
                        os.path.join(download_dir, torrent_file.files().file_path(i))
                        for i in range(0, torrent_file.files().num_files())
                    ]

                    if is_media:
                        from platypush.plugins.media import MediaPlugin

                        files = [f for f in files if MediaPlugin.is_video_file(f)]

                self.torrent_state[torrent]['download_rate'] = status.download_rate
                self.torrent_state[torrent]['name'] = status.name
                self.torrent_state[torrent]['num_peers'] = status.num_peers
                self.torrent_state[torrent]['paused'] = status.paused
                self.torrent_state[torrent]['progress'] = round(
                    100 * status.progress, 2
                )
                self.torrent_state[torrent]['state'] = status.state.name
                self.torrent_state[torrent]['title'] = status.name
                self.torrent_state[torrent]['torrent'] = torrent
                self.torrent_state[torrent]['upload_rate'] = status.upload_rate
                self.torrent_state[torrent]['url'] = torrent
                self.torrent_state[torrent]['files'] = files

                if transfer.has_metadata() and not metadata_downloaded:
                    self._fire_event(
                        TorrentDownloadedMetadataEvent(**self.torrent_state[torrent]),
                        event_hndl,
                    )
                    metadata_downloaded = True

                if status.state == status.downloading and not download_started:
                    self._fire_event(
                        TorrentDownloadStartEvent(**self.torrent_state[torrent]),
                        event_hndl,
                    )
                    download_started = True

                if last_status and status.progress != last_status.progress:
                    self._fire_event(
                        TorrentDownloadProgressEvent(**self.torrent_state[torrent]),
                        event_hndl,
                    )

                if not last_status or status.state != last_status.state:
                    self._fire_event(
                        TorrentStateChangeEvent(**self.torrent_state[torrent]),
                        event_hndl,
                    )

                if last_status and status.paused != last_status.paused:
                    if status.paused:
                        self._fire_event(
                            TorrentPausedEvent(**self.torrent_state[torrent]),
                            event_hndl,
                        )
                    else:
                        self._fire_event(
                            TorrentResumedEvent(**self.torrent_state[torrent]),
                            event_hndl,
                        )

                last_status = status
                time.sleep(self._MONITOR_CHECK_INTERVAL)

            if transfer and transfer.is_finished():
                self._fire_event(
                    TorrentDownloadCompletedEvent(**self.torrent_state[torrent]),
                    event_hndl,
                )

            self.remove(torrent)
            return files

        return thread

    def _get_session(self):
        if self._lt_session:
            return self._lt_session

        import libtorrent as lt

        self._lt_session = lt.session()
        return self._lt_session

    @action
    def download(
        self, torrent, download_dir=None, _async=False, event_hndl=None, is_media=False
    ):
        """
        Download a torrent.

        :param torrent: Torrent to download. Supported formats:

            * Magnet URLs
            * Torrent URLs
            * Local torrent file

        :type torrent: str

        :param download_dir: Directory to download, overrides the default download_dir attribute (default: None)
        :type download_dir: str

        :param _async: If true then the method will add the torrent to the transfer and then return. Updates on
            the download status should be retrieved either by listening to torrent events or registering the
            event handler. If false (default) then the method will exit only when the torrent download is complete.
        :type _async: bool

        :param event_hndl: A function that takes an event object as argument and is invoked upon a new torrent event
            (download started, progressing, completed etc.)
        :type event_hndl: function

        :param is_media: Set it to true if you're downloading a media file that you'd like to stream as soon as the
            first chunks are available. If so, then the events and the status method will only include media files
        :type is_media: bool
        """

        if torrent in self.torrent_state and torrent in self.transfers:
            return self.torrent_state[torrent]

        import libtorrent as lt

        if not download_dir:
            if self.download_dir:
                download_dir = self.download_dir
            else:
                raise RuntimeError('download_dir not specified')

        download_dir = os.path.abspath(os.path.expanduser(download_dir))
        os.makedirs(download_dir, exist_ok=True)
        info, file_info, torrent_file, magnet = self._get_torrent_info(
            torrent, download_dir
        )

        if torrent in self._sessions:
            self.logger.info('A torrent session is already running for %s', torrent)
            return self.torrent_state.get(torrent, {})

        session = self._get_session()
        session.listen_on(*self.torrent_ports)
        self._sessions[torrent] = session

        params = {
            'save_path': download_dir,
            'storage_mode': lt.storage_mode_t.storage_mode_sparse,
        }

        if magnet:
            transfer = lt.add_magnet_uri(session, magnet, params)
        else:
            params['ti'] = file_info
            transfer = session.add_torrent(params)

        self.transfers[torrent] = transfer
        self.torrent_state[torrent] = {
            'url': torrent,
            'title': transfer.status().name,
            'trackers': info['trackers'],
            'save_path': download_dir,
            'torrent_file': torrent_file,
        }

        self._fire_event(TorrentQueuedEvent(url=torrent), event_hndl)
        self.logger.info(
            'Downloading "%s" to "%s" from [%s]', info['name'], download_dir, torrent
        )
        monitor_thread = self._torrent_monitor(
            torrent=torrent,
            transfer=transfer,
            download_dir=download_dir,
            event_hndl=event_hndl,
            is_media=is_media,
        )

        if not _async:
            return monitor_thread()

        threading.Thread(target=monitor_thread).start()
        return self.torrent_state[torrent]

    @action
    def status(self, torrent=None):
        """
        Get the status of the current transfers.

        :param torrent: Torrent path, URL or magnet URI whose status will be retrieve (default: None, retrieve all
            current transfers)
        :type torrent: str

        :returns: A dictionary in the format torrent_url -> status
        """

        if torrent:
            return self.torrent_state.get(torrent)
        return self.torrent_state

    @action
    def pause(self, torrent):
        """
        Pause/resume a torrent transfer.

        :param torrent: Torrent URL as returned from `get_status()`
        :type torrent: str
        """

        if torrent not in self.transfers:
            return None, f"No transfer in progress for {torrent}"

        if self.torrent_state[torrent].get('paused', False):
            self.transfers[torrent].resume()
        else:
            self.transfers[torrent].pause()

        return self.torrent_state[torrent]

    @action
    def resume(self, torrent):
        """
        Resume a torrent transfer.

        :param torrent: Torrent URL as returned from `get_status()`
        :type torrent: str
        """

        assert torrent in self.transfers, f"No transfer in progress for {torrent}"
        self.transfers[torrent].resume()

    @action
    def remove(self, torrent):
        """
        Stops and removes a torrent transfer.

        :param torrent: Torrent URL as returned from `get_status()`
        :type torrent: str
        """

        if not self.transfers.get(torrent):
            self.logger.info('No transfer in progress for %s', torrent)
            return

        self.transfers[torrent].pause()
        del self.torrent_state[torrent]
        del self.transfers[torrent]
        if torrent in self._sessions:
            del self._sessions[torrent]

    @action
    def quit(self):
        """
        Quits all the transfers and the active session
        """
        transfers = self.transfers.copy()
        for torrent in transfers:
            self.remove(torrent)

    @staticmethod
    def _generate_rand_filename(length=16):
        name = ''
        for _ in range(0, length):
            name += hex(random.randint(0, 15))[2:].upper()
        return name + '.torrent'


# vim:sw=4:ts=4:et:



from multiprocessing import RLock, Queue, active_children
import os
from queue import Empty
import socket
import tempfile
from typing import Optional

from platypush.process import ControllableProcess

from ._base import Command
from ._reader import CommandReader
from ._writer import CommandWriter


class CommandStream(ControllableProcess):
    """
    The command stream is an abstraction built around a UNIX socket that allows
    the application to communicate commands to its controller.

    :param path: Path to the UNIX socket
    """

    _max_queue_size = 100
    """Maximum number of commands that can be queued in the stream."""

    _default_sock_path = tempfile.gettempdir() + '/platypush-cmd-stream.sock'
    """Default path to the UNIX socket."""

    _close_timeout = 5.0
    """Close the client socket after this amount of seconds."""

    def __init__(self, path: Optional[str] = None):
        super().__init__(name='platypush:cmd:stream')
        self.path = os.path.abspath(os.path.expanduser(path or self._default_sock_path))
        self._sock: Optional[socket.socket] = None
        self._cmd_queue = Queue()
        self._close_lock = RLock()

    def reset(self):
        if self._sock is not None:
            try:
                self._sock.close()
            except socket.error:
                pass

            self._sock = None

        try:
            os.unlink(self.path)
        except FileNotFoundError:
            pass

        self._cmd_queue.close()
        self._cmd_queue = Queue()

    def close(self) -> None:
        self.reset()
        return super().close()

    def _term_or_kill(self, kill: bool = False, visited=None) -> None:
        func = 'kill' if kill else 'terminate'
        visited = visited or set()
        visited.add(id(self))

        for child in active_children():
            child_id = id(child)
            if child_id not in visited:
                visited.add(child_id)
                getattr(child, func)()

        getattr(super(), func)()

    def terminate(self, visited=None) -> None:
        self._term_or_kill(kill=False, visited=visited)

    def kill(self) -> None:
        self._term_or_kill(kill=True)

    def __enter__(self) -> "CommandStream":
        self.reset()
        sock = self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.bind(self.path)
        os.chmod(self.path, 0o600)
        sock.listen(1)
        self.start()
        return self

    def __exit__(self, *_, **__):
        with self._close_lock:
            self.terminate()
            self.join(1)

            try:
                self.close()
            except Exception as e:
                self.logger.warning(
                    '%s on command stream close: %s', type(e).__name__, str(e)
                )

            try:
                self.kill()
            except Exception as e:
                self.logger.debug(
                    '%s on command stream kill: %s', type(e).__name__, str(e)
                )

    def _serve(self, sock: socket.socket):
        """
        Serves the command stream.

        :param sock: Client socket to serve
        """
        reader = CommandReader()
        cmd = reader.read(sock)
        if cmd:
            self._cmd_queue.put_nowait(cmd)

    def read(self, timeout: Optional[float] = None) -> Optional[Command]:
        """
        Reads commands from the command stream.

        :param timeout: Maximum time to wait for a command.
        :return: The command that was received, if any.
        """
        try:
            return self._cmd_queue.get(timeout=timeout)
        except Empty:
            return None

    def write(self, cmd: Command) -> None:
        """
        Writes a command to the command stream.

        :param cmd: Command to write
        :raise AssertionError: If the command cannot be written
        """
        sock = self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            sock.connect(self.path)
            self.logger.debug('Sending command: %s', cmd)
            CommandWriter().write(cmd, sock)
        except OSError as e:
            raise AssertionError(f'Unable to connect to socket {self.path}: {e}') from e
        finally:
            sock.close()

    @staticmethod
    def _close_client(sock: socket.socket):
        try:
            sock.close()
        except OSError:
            pass

    def main(self):
        while self._sock and not self.should_stop:
            sock = self._sock

            try:
                conn, _ = sock.accept()
            except ConnectionResetError:
                continue
            except KeyboardInterrupt:
                break

            try:
                self._serve(conn)
            except Exception as e:
                self.logger.warning('Unexpected socket error: %s', e, exc_info=True)
            finally:
                self._close_client(conn)



from datetime import datetime

from dateutil.tz import gettz
from marshmallow import pre_load

from .._base import SystemBaseSchema


class ProcessBaseSchema(SystemBaseSchema):
    """
    Base schema for system processes.
    """

    @pre_load
    def pre_load(self, data, **_) -> dict:
        if hasattr(data, 'as_dict'):
            data = data.as_dict()

        started_ts = data.pop('create_time', None)
        if started_ts is not None:
            data['started'] = datetime.fromtimestamp(started_ts).replace(tzinfo=gettz())

        data['command_line'] = data.pop('cmdline', None)
        data['cpu_percent'] = data.pop('cpu_percent') / 100
        data['current_directory'] = data.pop('cwd', None)
        data['memory_percent'] = data.pop('memory_percent') / 100
        data['parent_pid'] = data.pop('ppid', None)
        return data



import logging

from sqlalchemy import (
    Column,
    ForeignKey,
    Integer,
)

from platypush.common.db import is_defined

from .sensors import EnumSensor

logger = logging.getLogger(__name__)


if not is_defined('button'):

    class Button(EnumSensor):
        __tablename__ = 'button'

        id = Column(
            Integer, ForeignKey(EnumSensor.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from typing import Optional

from platypush.plugins.lcd import LcdPlugin


class LcdI2cPlugin(LcdPlugin):
    """
    Plugin to write to an LCD display connected via I2C.
    Adafruit I2C/SPI LCD Backback is supported.

    Warning: You might need a level shifter (that supports i2c) between the
    SCL/SDA connections on the MCP chip / backpack and the Raspberry Pi.

    Otherwise, you might damage the Pi and possibly any other 3.3V i2c devices
    connected on the i2c bus. Or cause reliability issues.

    The SCL/SDA are rated 0.7*VDD on the MCP23008, so it needs 3.5V on the
    SCL/SDA when 5V is applied to drive the LCD.

    The MCP23008 and MCP23017 needs to be connected exactly the same way as the
    backpack.

    For complete schematics see the adafruit page at:
    https://learn.adafruit.com/i2c-spi-lcd-backpack/

    4-bit operations. I2C only supported.

    Pin mapping::

        7  | 6  | 5  | 4  | 3  | 2 | 1  | 0
        BL | D7 | D6 | D5 | D4 | E | RS | -

    """

    def __init__(
        self,
        i2c_expander: str,
        address: int,
        expander_params: Optional[dict] = None,
        port: int = 1,
        cols: int = 16,
        rows: int = 2,
        backlight_enabled: bool = True,
        dotsize: int = 8,
        charmap: str = 'A02',
        auto_linebreaks: bool = True,
        **kwargs
    ):
        """
        :param i2c_expander: Set your I²C chip type. Supported: "PCF8574",
            "MCP23008", "MCP23017".
        :param address: The I2C address of your LCD.
        :param expander_params: Parameters for expanders, in a dictionary. Only
            needed for MCP23017 gpio_bank - This must be either ``A`` or ``B``.
            If you have a HAT, A is usually marked 1 and B is 2. Example:
            ``expander_params={'gpio_bank': 'A'}``
        :param port: The I2C port number. Default: ``1``.
        :param cols: Number of columns per row (usually 16 or 20). Default: ``16``.
        :param rows: Number of display rows (usually 1, 2 or 4). Default: ``2``.
        :param backlight_enabled: Whether the backlight is enabled initially.
            Default: ``True``. Has no effect if pin_backlight is ``None``
        :param dotsize: Some 1 line displays allow a font height of 10px.
            Allowed: ``8`` or ``10``. Default: ``8``.
        :param charmap: The character map used. Depends on your LCD. This must
            be either ``A00`` or ``A02`` or ``ST0B``. Default: ``A02``.
        :param auto_linebreaks: Whether or not to automatically insert line
            breaks. Default: ``True``.
        """
        super().__init__(**kwargs)

        self.i2c_expander = i2c_expander
        self.address = address
        self.expander_params = expander_params or {}
        self.port = port
        self.cols = cols
        self.rows = rows
        self.backlight_enabled = backlight_enabled
        self.dotsize = dotsize
        self.auto_linebreaks = auto_linebreaks
        self.charmap = charmap

    def _get_lcd(self):
        from RPLCD.i2c import CharLCD

        return CharLCD(
            cols=self.cols,
            rows=self.rows,
            i2c_expander=self.i2c_expander,
            address=self.address,
            port=self.port,
            backlight_enabled=self.backlight_enabled,
            dotsize=self.dotsize,
            charmap=self.charmap,
            auto_linebreaks=self.auto_linebreaks,
        )


class LcdI2CPlugin(LcdI2cPlugin):
    pass


# vim:sw=4:ts=4:et:



import sys

from ._runner import ApplicationRunner


def main():
    """
    Main application entry point.

    This is usually the entry point that you want to use to start your
    application, rather than :meth:`platypush.app.main`, as this entry point
    wraps the main application in a controllable process.
    """

    app_runner = ApplicationRunner()
    app_runner.run(*sys.argv[1:])


__all__ = ["ApplicationRunner", "main"]



import json
import logging
import re
import time
from dataclasses import asdict
from queue import Empty
from threading import Event, RLock, Thread
from typing import Dict, Generator, List, Optional, Type

import websocket

from platypush.context import get_bus
from platypush.message.event.music import (
    MusicEvent,
    MusicPauseEvent,
    MusicPlayEvent,
    MusicStopEvent,
    MuteChangeEvent,
    NewPlayingTrackEvent,
    PlaybackConsumeModeChangeEvent,
    PlaybackRandomModeChangeEvent,
    PlaybackRepeatModeChangeEvent,
    PlaybackSingleModeChangeEvent,
    PlaylistChangeEvent,
    SeekChangeEvent,
    VolumeChangeEvent,
)
from platypush.plugins.media import PlayerState

from ._common import DEFAULT_TIMEOUT
from ._conf import MopidyConfig
from ._status import MopidyStatus
from ._sync import PlaylistSync
from ._task import MopidyTask
from ._track import MopidyTrack


class MopidyClient(Thread):
    """
    Thread that listens for Mopidy events and posts them to the bus.
    """

    def __init__(
        self,
        config: MopidyConfig,
        status: MopidyStatus,
        stop_event: Event,
        playlist_sync: PlaylistSync,
        tasks: Dict[int, MopidyTask],
        **_,
    ):
        super().__init__(name='platypush:mopidy:listener')

        self.logger = logging.getLogger('platypush:mopidy:listener')
        self.config = config
        self._status = status
        self._stop_event = stop_event
        self._playlist_sync = playlist_sync
        self._tasks = tasks
        self._refresh_in_progress = Event()
        self._refresh_lock = RLock()
        self._req_lock = RLock()
        self._close_lock = RLock()
        self._tracks: List[MopidyTrack] = []
        self._msg_id = 0
        self._ws = None
        self.connected_event = Event()
        self.closed_event = Event()

    @property
    def _bus(self):
        return get_bus()

    @property
    def status(self):
        return self._status

    @property
    def tracks(self):
        return self._tracks

    def should_stop(self):
        return self._stop_event.is_set()

    def wait_stop(self, timeout: Optional[float] = None):
        self._stop_event.wait(timeout=timeout)

    def make_task(self, method: str, **args: dict) -> MopidyTask:
        with self._req_lock:
            self._msg_id += 1
            task = MopidyTask(
                id=self._msg_id,
                method=method,
                args=args or {},
            )

        self._tasks[task.id] = task
        return task

    def send(self, *tasks: MopidyTask):
        """
        Send a list of tasks to the Mopidy server.
        """
        assert self._ws, 'Websocket not connected'

        for task in tasks:
            with self._req_lock:
                task.send(self._ws)

    def gather(
        self,
        *tasks: MopidyTask,
        timeout: Optional[float] = DEFAULT_TIMEOUT,
    ) -> Generator:
        t_start = time.time()

        for task in tasks:
            remaining_timeout = (
                max(0, timeout - (time.time() - t_start)) if timeout else None
            )

            if not self._tasks.get(task.id):
                yield None

            try:
                ret = self._tasks[task.id].get_response(timeout=remaining_timeout)
                assert not isinstance(ret, Exception), ret
                self.logger.debug('Got response for %s: %s', task, ret)
                yield ret
            except Empty as e:
                t = self._tasks.get(task.id)
                err = 'Mopidy request timeout'
                if t:
                    err += f' - method: {t.method} args: {t.args}'

                raise TimeoutError(err) from e
            finally:
                self._tasks.pop(task.id, None)

    def exec(self, *msgs: dict, timeout: Optional[float] = DEFAULT_TIMEOUT) -> list:
        tasks = [self.make_task(**msg) for msg in msgs]
        for task in tasks:
            self.send(task)

        return list(self.gather(*tasks, timeout=timeout))

    def refresh_status(  # pylint: disable=too-many-branches
        self, timeout: Optional[float] = DEFAULT_TIMEOUT, with_tracks: bool = False
    ):
        if self._refresh_in_progress.is_set():
            return

        # Unless we are also retrieving the tracks, we don't need a high timeout here.
        # A high timeout may result in event lagging if some of the responses aren't received.
        # In this case, better to timeout early and retry.
        if not with_tracks:
            timeout = 5

        events = []

        try:
            with self._refresh_lock:
                self._refresh_in_progress.set()
                # Refresh the tracklist attributes
                opts = ('repeat', 'random', 'single', 'consume')
                ret = self.exec(
                    *[
                        *[{'method': f'core.tracklist.get_{opt}'} for opt in opts],
                        {'method': 'core.playback.get_current_tl_track'},
                        {'method': 'core.playback.get_state'},
                        {'method': 'core.mixer.get_volume'},
                        {'method': 'core.playback.get_time_position'},
                        *(
                            [{'method': 'core.tracklist.get_tl_tracks'}]
                            if with_tracks
                            else []
                        ),
                    ],
                    timeout=timeout,
                )

                for i, opt in enumerate(opts):
                    new_value = ret[i]
                    if opt == 'random' and self._status.random != new_value:
                        events.append(
                            (PlaybackRandomModeChangeEvent, {'state': new_value})
                        )
                    if opt == 'repeat' and self._status.repeat != new_value:
                        events.append(
                            (PlaybackRepeatModeChangeEvent, {'state': new_value})
                        )
                    if opt == 'single' and self._status.single != new_value:
                        events.append(
                            (PlaybackSingleModeChangeEvent, {'state': new_value})
                        )
                    if opt == 'consume' and self._status.consume != new_value:
                        events.append(
                            (PlaybackConsumeModeChangeEvent, {'state': new_value})
                        )

                    setattr(self._status, opt, new_value)

                # Get remaining info
                track = MopidyTrack.parse(ret[4])
                state, volume, t = ret[5:8]

                if track:
                    idx = self.exec(
                        {
                            'method': 'core.tracklist.index',
                            'tlid': track.track_id,
                        },
                        timeout=timeout,
                    )[0]

                    self._status.track = track
                    self._status.duration = track.time
                    if idx is not None:
                        self._status.playing_pos = self._status.track.playlist_pos = idx

                    if track != self._status.track and state != 'stopped':
                        events.append((NewPlayingTrackEvent, {}))

                if state != self._status.state:
                    if state == 'paused':
                        self._status.state = PlayerState.PAUSE
                        events.append((MusicPauseEvent, {}))
                    elif state == 'playing':
                        self._status.state = PlayerState.PLAY
                        events.append((MusicPlayEvent, {}))
                    elif state == 'stopped':
                        self._status.state = PlayerState.STOP
                        events.append((MusicStopEvent, {}))

                if volume != self._status.volume:
                    self._status.volume = volume
                    events.append((VolumeChangeEvent, {'volume': volume}))

                if t != self._status.time:
                    self._status.time = t / 1000
                    events.append((SeekChangeEvent, {'position': self._status.time}))

                if with_tracks:
                    self._tracks = [  # type: ignore
                        MopidyTrack.parse({**t, 'playlist_pos': i})
                        for i, t in enumerate(ret[8])
                    ]

            for evt in events:
                self._post_event(evt[0], **evt[1])
        finally:
            self._refresh_in_progress.clear()

    def _refresh_status(
        self, timeout: Optional[float] = DEFAULT_TIMEOUT, with_tracks: bool = False
    ):
        """
        Refresh the status from the Mopidy server.

        It runs in a separate thread because the status refresh logic runs in
        synchronous mode, and it would block the main thread preventing the
        listener from receiving new messages.

        Also, an event+reenrant lock mechanism is used to ensure that only one
        refresh task is running at a time.
        """
        if self._refresh_in_progress.is_set():
            return

        with self._refresh_lock:
            Thread(
                target=self.refresh_status,
                kwargs={'timeout': timeout, 'with_tracks': with_tracks},
                daemon=True,
            ).start()

    def _post_event(self, evt_cls: Type[MusicEvent], **kwargs):
        self._bus.post(
            evt_cls(
                status=asdict(self._status),
                track=asdict(self._status.track) if self._status.track else None,
                plugin_name='music.mopidy',
                **kwargs,
            )
        )

    def _handle_error(self, msg: dict):
        msg_id = msg.get('id')
        err = msg.get('error')
        if not err:
            return

        err_data = err.get('data', {})
        tb = err_data.get('traceback')
        self.logger.warning(
            'Mopidy error: %s: %s: %s',
            err.get('message'),
            err_data.get('type'),
            err_data.get('message'),
        )
        if tb:
            self.logger.warning(tb)

        if msg_id:
            task = self._tasks.get(msg_id)
            if task:
                task.put_response(
                    RuntimeError(err.get('message') + ': ' + err_data.get('message'))
                )

    def on_pause(self, *_, **__):
        self._status.state = PlayerState.PAUSE
        self._post_event(MusicPauseEvent)

    def on_resume(self, *_, **__):
        self._status.state = PlayerState.PLAY
        self._post_event(MusicPlayEvent)

    def on_start(self, *_, **__):
        self._refresh_status()

    def on_end(self, *_, **__):
        self._refresh_status()

    def on_state_change(self, msg: dict, *_, **__):
        state = msg.get('new_state')
        if state == 'playing':
            self._status.state = PlayerState.PLAY
            self._post_event(MusicPlayEvent)
        elif state == 'paused':
            self._status.state = PlayerState.PAUSE
            self._post_event(MusicPauseEvent)
        elif state == 'stopped':
            self._status.state = PlayerState.STOP
            self._post_event(MusicStopEvent)

    def on_title_change(self, msg: dict, *_, track: MopidyTrack, **__):
        title = msg.get('title', '')
        m = re.match(r'^\s*(.+?)\s+-\s+(.*)\s*$', title)
        if not m:
            return

        track.artist = m.group(1)
        track.title = m.group(2)
        self._post_event(NewPlayingTrackEvent)

    def on_volume_change(self, msg: dict, *_, **__):
        volume = msg.get('volume')
        if volume is None or volume == self._status.volume:
            return

        self._status.volume = volume
        self._post_event(VolumeChangeEvent, volume=volume)

    def on_mute_change(self, msg: dict, *_, **__):
        mute = msg.get('mute')
        if mute is None or mute == self._status.mute:
            return

        self._status.mute = mute
        self._post_event(MuteChangeEvent, mute=mute)

    def on_seek(self, msg: dict, *_, **__):
        position = msg.get('time_position')
        if position is None:
            return

        t = position / 1000
        if t == self._status.time:
            return

        self._status.time = t
        self._post_event(SeekChangeEvent, position=self._status.time)

    def on_tracklist_change(self, *_, **__):
        should_proceed = self._playlist_sync.wait_for_loading(timeout=2)
        if not should_proceed:
            return

        self.logger.debug('Tracklist changed, refreshing changes')
        self._refresh_status(with_tracks=True)
        self._post_event(PlaylistChangeEvent)

    def on_options_change(self, *_, **__):
        self._refresh_status()

    def _on_msg(self, *args):
        msg = args[1] if len(args) > 1 else args[0]
        msg = json.loads(msg)
        msg_id = msg.get('id')
        event = msg.get('event')
        track: Optional[MopidyTrack] = None
        self.logger.debug('Received Mopidy message: %s', msg)

        if msg.get('error'):
            self._handle_error(msg)
            return

        if msg_id:
            task = self._tasks.get(msg_id)
            if task:
                task.put_response(msg)
                return

        if not event:
            return

        if msg.get('tl_track'):
            track = MopidyTrack.parse(msg['tl_track'])
            is_new_track = track and track != self._status.track
            self._status.track = track
            if is_new_track:
                self._post_event(NewPlayingTrackEvent)

        hndl = self._msg_handlers.get(event)
        if not hndl:
            return

        hndl(self, msg, track=track)

    def _on_error(self, *args):
        error = args[1] if len(args) > 1 else args[0]
        ws = args[0] if len(args) > 1 else None
        self.logger.warning('Mopidy websocket error: %s', error)
        if ws:
            ws.close()

    def _on_close(self, *_):
        self.connected_event.clear()
        self.closed_event.set()

        if self._ws:
            try:
                self._ws.close()
            except Exception as e:
                self.logger.debug(e, exc_info=True)
            finally:
                self._ws = None

        self.logger.warning('Mopidy websocket connection closed')

    def _on_open(self, *_):
        self.connected_event.set()
        self.closed_event.clear()
        self.logger.info('Mopidy websocket connected')
        self._refresh_status(with_tracks=True)

    def _connect(self):
        if not self._ws:
            self._ws = websocket.WebSocketApp(
                self.config.url,
                on_open=self._on_open,
                on_message=self._on_msg,
                on_error=self._on_error,
                on_close=self._on_close,
            )

            self._ws.run_forever()

    def run(self):
        while not self.should_stop():
            try:
                self._connect()
            except Exception as e:
                self.logger.warning(
                    'Error on websocket connection: %s', e, exc_info=True
                )
            finally:
                self.connected_event.clear()
                self.closed_event.set()
                self.wait_stop(10)

    def stop(self):
        with self._close_lock:
            if self._ws:
                self._ws.close()
                self._ws = None

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.stop()

    _msg_handlers = {
        'track_playback_paused': on_pause,
        'playback_state_changed': on_state_change,
        'track_playback_resumed': on_resume,
        'track_playback_ended': on_end,
        'track_playback_started': on_start,
        'stream_title_changed': on_title_change,
        'volume_changed': on_volume_change,
        'mute_changed': on_mute_change,
        'seeked': on_seek,
        'tracklist_changed': on_tracklist_change,
        'options_changed': on_options_change,
    }


# vim:sw=4:ts=4:et:



from dataclasses import dataclass
from enum import Enum


class ConnectedState(Enum):
    """
    Enum to represent the connection state of a joystick.
    """

    UNKNOWN = 0
    CONNECTED = 1
    DISCONNECTED = 2


@dataclass
class JoystickState:
    """
    Dataclass that holds the joystick state properties.
    """

    left_joystick_y: float = 0
    left_joystick_x: float = 0
    right_joystick_y: float = 0
    right_joystick_x: float = 0
    left_trigger: float = 0
    right_trigger: float = 0
    left_bumper: int = 0
    right_bumper: int = 0
    a: int = 0
    x: int = 0
    y: int = 0
    b: int = 0
    left_thumb: int = 0
    right_thumb: int = 0
    back: int = 0
    start: int = 0
    left_dir_pad: int = 0
    right_dir_pad: int = 0
    up_dir_pad: int = 0
    down_dir_pad: int = 0


@dataclass
class JoystickDeviceState:
    """
    Dataclass that holds the joystick device state properties.
    """

    device: str
    state: JoystickState



"""Added password hashing parameters to user table

Revision ID: 0876439530cb
Revises: c39ac404119b
Create Date: 2024-05-05 20:57:02.820575

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0876439530cb'
down_revision = 'c39ac404119b'
branch_labels = None
depends_on = None


def upgrade() -> None:
    conn = op.get_bind()
    metadata = sa.MetaData()
    metadata.reflect(bind=conn)
    user_table = metadata.tables.get('user')

    if user_table is None:
        print('The table `user` does not exist, skipping migration')
        return

    if 'password_salt' not in user_table.columns:
        op.add_column('user', sa.Column('password_salt', sa.String(), nullable=True))

    if 'hmac_iterations' not in user_table.columns:
        op.add_column('user', sa.Column('hmac_iterations', sa.Integer(), nullable=True))


def downgrade() -> None:
    conn = op.get_bind()
    metadata = sa.MetaData()
    metadata.reflect(bind=conn)
    user_table = metadata.tables.get('user')

    if user_table is None:
        print('The table `user` does not exist, skipping migration')
        return

    if 'password_salt' in user_table.columns:
        op.drop_column('user', 'password_salt')

    if 'hmac_iterations' in user_table.columns:
        op.drop_column('user', 'hmac_iterations')



from queue import Empty
import wx

from platypush.plugins.camera.model.writer.preview import PreviewWriter


class Panel(wx.Panel):
    def __init__(self, parent, process, width: int, height: int):
        import wx
        super().__init__(parent, -1)

        self.process: PreviewWriter = process
        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        self.SetSize(width, height)
        self.Bind(wx.EVT_PAINT, self.on_paint)
        self.update()

    @staticmethod
    def img_to_bitmap(image) -> wx.Bitmap:
        import wx
        return wx.Bitmap.FromBuffer(image.width, image.height, image.tobytes())

    def get_bitmap(self):
        try:
            return self.process.bitmap_queue.get(block=True, timeout=1.0)
        except Empty:
            return None

    def update(self):
        import wx
        self.Refresh()
        self.Update()
        wx.CallLater(15, self.update)

    def create_bitmap(self):
        image = self.get_bitmap()
        if image is None:
            return

        return self.img_to_bitmap(image)

    def on_paint(self, *_, **__):
        import wx
        bitmap = self.create_bitmap()
        if not bitmap:
            return

        dc = wx.AutoBufferedPaintDC(self)
        dc.DrawBitmap(bitmap, 0, 0)


class Frame(wx.Frame):
    def __init__(self, process):
        import wx
        style = wx.DEFAULT_FRAME_STYLE & ~wx.RESIZE_BORDER & ~wx.MAXIMIZE_BOX
        self.process = process
        image = self.process.bitmap_queue.get()

        super().__init__(None, -1, process.camera.info.device or 'Camera Preview', style=style)
        self.Bind(wx.EVT_WINDOW_DESTROY, self.on_close)
        self.panel = Panel(self, process, width=image.width, height=image.height)
        self.Fit()

    def on_close(self, *_, **__):
        self.process.close()


# vim:sw=4:ts=4:et:



from io import BytesIO
from dataclasses import dataclass, field
from typing import List

import numpy as np
from pydub import AudioSegment, silence

from platypush.common.assistant import AudioFrame


@dataclass
class RecordingState:
    """
    Current state of the audio recording.
    """

    sample_rate: int
    channels: int
    min_silence_secs: float
    silence_threshold: int
    silence_duration: float = 0.0
    audio_segments: List[AudioSegment] = field(default_factory=list)
    duration: float = 0.0
    conversation_started: bool = False

    def _silence_duration(self, audio: AudioSegment) -> float:
        silent_frames = [
            (start / 1000, stop / 1000)
            for start, stop in silence.detect_silence(
                audio,
                min_silence_len=int(self.min_silence_secs * 1000),
                silence_thresh=int(self.silence_threshold),
            )
        ]

        return sum(stop - start for start, stop in silent_frames)

    def _to_audio_segment(self, data: np.ndarray) -> AudioSegment:
        return AudioSegment(
            data.tobytes(),
            frame_rate=self.sample_rate,
            sample_width=data.dtype.itemsize,
            channels=self.channels,
        )

    def _add_audio_segment(self, audio: AudioSegment):
        self.audio_segments.append(audio)
        self.duration += audio.duration_seconds
        silence_duration = self._silence_duration(audio)
        is_mostly_silent = silence_duration >= audio.duration_seconds * 0.75

        if is_mostly_silent:
            self.silence_duration += silence_duration
        else:
            self.conversation_started = True
            self.silence_duration = 0.0

    def is_silent(self) -> bool:
        return self.silence_duration >= self.duration

    def add_audio(self, audio: AudioFrame):
        self._add_audio_segment(self._to_audio_segment(audio.data))

    def export_audio(self) -> BytesIO:
        buffer = BytesIO()
        if not self.audio_segments:
            return buffer

        audio = self.audio_segments[0]
        for segment in self.audio_segments[1:]:
            audio += segment

        audio.export(buffer, format="mp3", bitrate='92')
        return buffer

    def reset(self):
        self.audio_segments.clear()
        self.duration = 0.0
        self.silence_duration = 0.0
        self.conversation_started = False



import logging
import os
import signal
import sys
from threading import Thread
from typing import Optional

from platypush.cli import parse_cmdline
from platypush.commands import CommandStream

from ._app import ApplicationProcess


class ApplicationRunner:
    """
    Runner for the main application.

    It wraps the main application and provides an interface to control it
    externally.
    """

    _default_timeout = 0.5

    def __init__(self):
        logging.basicConfig(level=logging.INFO, stream=sys.stdout)
        self.logger = logging.getLogger('platypush:runner')
        self._proc: Optional[ApplicationProcess] = None
        self._stream: Optional[CommandStream] = None

    def _listen(self, stream: CommandStream):
        """
        Listens for external application commands and executes them.
        """
        while self.is_running:
            cmd = stream.read(timeout=self._default_timeout)
            if not cmd:
                continue

            self.logger.info(cmd)
            Thread(target=cmd, args=(self,)).start()

    def _print_version(self):
        from platypush import __version__

        print(__version__)
        sys.exit(0)

    def _run(self, *args: str) -> None:
        parsed_args = parse_cmdline(args)

        if parsed_args.version:
            self._print_version()

        signal.signal(signal.SIGTERM, lambda *_: self.stop())

        while True:
            with CommandStream(
                parsed_args.ctrl_sock
            ) as self._stream, ApplicationProcess(
                *args, pidfile=parsed_args.pidfile, timeout=self._default_timeout
            ) as self._proc:
                try:
                    self._listen(self._stream)
                except KeyboardInterrupt:
                    pass

                if self.should_restart:
                    self.logger.info('Restarting application...')
                    continue

                break

        self._stream = None

    def run(self, *args: str) -> None:
        try:
            self._run(*args)
        finally:
            self._proc = None

    def stop(self):
        if self._proc is not None:
            self._proc.stop()

        if self._stream and self._stream.pid:
            os.kill(self._stream.pid, signal.SIGKILL)
            self._stream = None

    def restart(self):
        if self._proc is not None:
            self._proc.mark_for_restart()

        self.stop()

    @property
    def is_running(self) -> bool:
        return bool(self._proc and self._proc.is_alive())

    @property
    def should_restart(self) -> bool:
        return self._proc.should_restart if self._proc is not None else False
